# samtools_utils.py
import os
import shutil
import subprocess
import tempfile
from dataclasses import dataclass
from typing import Callable, Dict, Optional

try:
    from plantvarfilter.linux import resolve_tool
except Exception:
    resolve_tool = None

LogFn = Callable[[str], None]


class SamtoolsError(RuntimeError):
    """Exception raised for samtools-related failures."""
    pass


def _resolve_exe(name: str) -> str:
    """Resolve executable path from bundled tools first, then PATH."""
    if resolve_tool is not None:
        exe = resolve_tool(name)
        if exe:
            return exe
    path = shutil.which(name)
    if path:
        return path
    raise SamtoolsError(
        f"Executable not found: {name}. Bundle it under PlantVarFilter/linux or add to PATH."
    )


def _run(cmd: list, log: Optional[LogFn] = None, cwd: Optional[str] = None) -> str:
    """Run a command, return stdout (decoded safely)."""
    if log:
        log(" ".join(cmd))
    proc = subprocess.run(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = proc.stdout.decode("utf-8", errors="replace")
    if proc.returncode != 0:
        raise SamtoolsError(f"Command failed ({proc.returncode}): {' '.join(cmd)}\n{out}")
    if out.strip() and log:
        for line in out.strip().splitlines():
            log(line)
    return out


def _run_to_file(cmd: list, out_path: str, log: Optional[LogFn] = None, cwd: Optional[str] = None):
    """Run a command and write its stdout directly to a file."""
    if log:
        log(" ".join(cmd) + f" > {out_path}")
    with open(out_path, "w") as fh:
        proc = subprocess.run(cmd, cwd=cwd, stdout=fh, stderr=subprocess.PIPE, text=True)
    if proc.returncode != 0:
        raise SamtoolsError(
            f"Command failed ({proc.returncode}): {' '.join(cmd)}\nSTDERR:\n{proc.stderr}"
        )


@dataclass
class PreprocessOutputs:
    final_bam: str
    bai: Optional[str]
    stats_files: Dict[str, str]  # keys: flagstat, stats, idxstats, depth, markdup


class Samtools:
    def __init__(self, exe: str = "samtools"):
        self.exe = _resolve_exe(exe)

    # -------- Basic ops --------
    def version(self) -> str:
        try:
            proc = subprocess.run([self.exe, "--version"],
                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            out = proc.stdout.decode("utf-8", errors="replace")
            return out.splitlines()[0].strip() if out else "samtools (unknown)"
        except Exception as e:
            return f"samtools (unknown): {e}"

    def sort_name(self, in_bam: str, out_bam: str, threads: int = 4, log: Optional[LogFn] = None):
        _run([self.exe, "sort", "-n", "-@", str(threads), "-o", out_bam, in_bam], log=log)

    def sort_coord(self, in_bam: str, out_bam: str, threads: int = 4,
                   tmpdir: Optional[str] = None, log: Optional[LogFn] = None):
        args = [self.exe, "sort", "-@", str(threads)]
        if tmpdir:
            args += ["-T", os.path.join(tmpdir, "samtools_sort_tmp")]
        args += ["-o", out_bam, in_bam]
        _run(args, log=log)

    def fixmate(self, in_bam_namesorted: str, out_bam: str, add_ms_tag: bool = True,
                log: Optional[LogFn] = None):
        args = [self.exe, "fixmate"]
        if add_ms_tag:
            args.append("-m")
        args += [in_bam_namesorted, out_bam]
        _run(args, log=log)

    def markdup(self, in_bam_coordsorted: str, out_bam: str, threads: int = 4,
                remove_dups: bool = False, stats_file: Optional[str] = None,
                log: Optional[LogFn] = None):
        args = [self.exe, "markdup", "-@", str(threads)]
        if remove_dups:
            args.append("-r")
        if stats_file:
            args += ["-f", stats_file]
        args += [in_bam_coordsorted, out_bam]
        _run(args, log=log)

    def index(self, bam: str, threads: int = 2, log: Optional[LogFn] = None) -> str:
        _run([self.exe, "index", "-@", str(threads), bam], log=log)
        # Typical outputs are bam.bai or basename.bai
        candidates = [bam + ".bai", os.path.splitext(bam)[0] + ".bai"]
        for p in candidates:
            if os.path.exists(p):
                return p
        return ""

    # -------- QC helpers --------
    def flagstat(self, bam: str, out_txt: str, log: Optional[LogFn] = None):
        _run_to_file([self.exe, "flagstat", bam], out_txt, log=log)

    def stats(self, bam: str, out_txt: str, log: Optional[LogFn] = None):
        _run_to_file([self.exe, "stats", bam], out_txt, log=log)

    def idxstats(self, bam: str, out_txt: str, log: Optional[LogFn] = None):
        _run_to_file([self.exe, "idxstats", bam], out_txt, log=log)

    def depth_all(self, bam: str, out_txt: str, max_depth: Optional[int] = None,
                  log: Optional[LogFn] = None):
        args = [self.exe, "depth", "-a"]
        if max_depth is not None:
            args += ["-d", str(max_depth)]
        args.append(bam)
        _run_to_file(args, out_txt, log=log)

    # -------- High-level pipeline --------
    def preprocess(self,
                   input_path: str,
                   out_prefix: Optional[str] = None,
                   threads: int = 4,
                   remove_dups: bool = False,
                   compute_stats: bool = True,
                   log: Optional[LogFn] = None,
                   keep_temps: bool = False) -> PreprocessOutputs:
        """
        Pipeline:
          0) If input is SAM/CRAM -> convert to BAM
          1) sort -n
          2) fixmate -m
          3) sort (coordinate)
          4) markdup [-r]
          5) index
          6) optional: flagstat, stats, idxstats, depth
        """
        if not os.path.exists(input_path):
            raise SamtoolsError(f"Input file not found: {input_path}")

        workdir = os.path.dirname(os.path.abspath(input_path))
        stem = os.path.splitext(os.path.basename(input_path))[0]
        base = out_prefix or f"{stem}.sm"
        base = os.path.join(workdir, base)

        tmpdir = tempfile.mkdtemp(prefix="samtools_", dir=workdir)
        try:
            # Step 0: ensure BAM
            ext = os.path.splitext(input_path)[1].lower()
            start_bam = base + ".start.bam"
            if ext == ".bam":
                if os.path.abspath(input_path) != os.path.abspath(start_bam):
                    shutil.copyfile(input_path, start_bam)
            elif ext == ".sam":
                _run([self.exe, "view", "-bS", "-o", start_bam, input_path], log=log)
            elif ext == ".cram":
                _run([self.exe, "view", "-b", "-o", start_bam, input_path], log=log)
            else:
                raise SamtoolsError(f"Unsupported input extension: {ext}")

            ns_bam = base + ".namesort.bam"
            fx_bam = base + ".fixmate.bam"
            cs_bam = base + ".coordsort.bam"
            dd_bam = base + ".dedup.bam"
            md_stats = base + ".markdup.txt"

            # 1) name sort
            self.sort_name(start_bam, ns_bam, threads=threads, log=log)
            # 2) fixmate
            self.fixmate(ns_bam, fx_bam, add_ms_tag=True, log=log)
            # 3) coordinate sort
            self.sort_coord(fx_bam, cs_bam, threads=threads, tmpdir=tmpdir, log=log)
            # 4) markdup
            self.markdup(cs_bam, dd_bam, threads=threads, remove_dups=remove_dups,
                         stats_file=md_stats, log=log)
            # 5) index
            bai = self.index(dd_bam, threads=max(1, threads // 2), log=log)

            stats_files: Dict[str, str] = {"markdup": md_stats}
            if compute_stats:
                flagstat_p = base + ".flagstat.txt"
                stats_p = base + ".stats.txt"
                idxstats_p = base + ".idxstats.txt"
                depth_p = base + ".depth.txt"
                self.flagstat(dd_bam, flagstat_p, log=log)
                self.stats(dd_bam, stats_p, log=log)
                self.idxstats(dd_bam, idxstats_p, log=log)
                self.depth_all(dd_bam, depth_p, log=log)
                stats_files.update({
                    "flagstat": flagstat_p,
                    "stats": stats_p,
                    "idxstats": idxstats_p,
                    "depth": depth_p
                })

            if not keep_temps:
                for p in (start_bam, ns_bam, fx_bam, cs_bam):
                    try:
                        if os.path.exists(p):
                            os.remove(p)
                    except Exception:
                        pass

            return PreprocessOutputs(final_bam=dd_bam, bai=bai, stats_files=stats_files)
        finally:
            try:
                os.rmdir(tmpdir)
            except Exception:
                pass



# annotation_utils.py
import os
import gzip
import io
from typing import Dict, Tuple, Optional

import numpy as np
import pandas as pd


class Annotator:
    """
    Lightweight GWAS annotator:
      - Loads GTF/GFF (optionally .gz)
      - Extracts 'gene' features with gene_id / gene_name
      - Builds per-chromosome sorted arrays for fast nearest-gene lookup
      - Annotates GWAS table with nearest gene within a user window (kb)
    """

    def __init__(self):
        self.genes_df: Optional[pd.DataFrame] = None
        self.index: Dict[str, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = {}

    # ---------- I/O helpers ----------
    @staticmethod
    def _open_any(path: str):
        if str(path).endswith(".gz"):
            with gzip.open(path, "rb") as fh:
                data = fh.read()
            return io.StringIO(data.decode("utf-8", errors="replace"))
        return open(path, "r", encoding="utf-8", errors="replace")

    @staticmethod
    def _parse_attrs(attr: str) -> Dict[str, str]:
        out = {}
        if not isinstance(attr, str):
            return out

        # GTF style: key "value"; key "value";
        # GFF style: key=value;key2=value2
        parts = []
        # try GFF k=v
        if "=" in attr and ";" in attr:
            parts = [p for p in attr.strip().split(";") if p.strip()]
            for p in parts:
                if "=" in p:
                    k, v = p.split("=", 1)
                    out[k.strip()] = v.strip().strip('"')
        else:
            # fall back to GTF
            semi = [p for p in attr.strip().split(";") if p.strip()]
            for p in semi:
                tokens = p.strip().split()
                if len(tokens) >= 2:
                    k = tokens[0].strip()
                    v = " ".join(tokens[1:]).strip().strip('"')
                    out[k] = v
        return out

    # ---------- GTF/GFF load ----------
    def load_gtf_or_gff(self, path: str) -> pd.DataFrame:
        cols = ["seqname", "source", "feature", "start", "end", "score", "strand", "frame", "attribute"]
        rows = []
        with self._open_any(path) as fh:
            for ln in fh:
                if not ln or ln.startswith("#"):
                    continue
                parts = ln.rstrip("\n").split("\t")
                if len(parts) < 9:
                    continue
                seq, src, feat, st, en, sc, strand, frame, attr = parts[:9]
                if feat.lower() != "gene":
                    continue
                try:
                    st_i = int(st)
                    en_i = int(en)
                except Exception:
                    continue
                a = self._parse_attrs(attr)
                gene_id = a.get("gene_id") or a.get("ID") or a.get("GeneID") or ""
                gene_name = a.get("gene_name") or a.get("Name") or a.get("gene") or gene_id
                rows.append((seq, st_i, en_i, gene_id, gene_name, strand))

        if not rows:
            raise ValueError("No 'gene' features found in supplied GTF/GFF.")

        df = pd.DataFrame(rows, columns=["Chr", "Start", "End", "GeneID", "GeneName", "Strand"])
        # Normalize chromosome naming (remove 'chr' prefix if mixed)
        if df["Chr"].str.startswith("chr").any():
            df["Chr"] = df["Chr"].str.replace("^chr", "", regex=True)
        self.genes_df = df
        return df

    # ---------- Index build ----------
    def build_index(self):
        if self.genes_df is None or self.genes_df.empty:
            raise RuntimeError("Call load_gtf_or_gff first.")

        self.index.clear()
        for chrom, sub in self.genes_df.groupby("Chr", sort=False):
            # sort by Start for binary search
            sub = sub.sort_values("Start").reset_index(drop=True)
            starts = sub["Start"].to_numpy(dtype=np.int64)
            ends = sub["End"].to_numpy(dtype=np.int64)
            names = sub["GeneName"].astype(str).to_numpy()
            ids = sub["GeneID"].astype(str).to_numpy()
            self.index[str(chrom)] = (starts, ends, names, ids)

    # ---------- nearest gene ----------
    @staticmethod
    def _nearest_distance(pos: int, starts: np.ndarray, ends: np.ndarray) -> Tuple[int, int]:
        """
        Returns (idx, distance_bp). If pos is inside a gene interval, distance=0.
        Otherwise min distance to closest interval edge. idx is index into arrays.
        """
        if starts.size == 0:
            return -1, np.iinfo(np.int64).max

        # locate insertion point
        i = np.searchsorted(starts, pos, side="left")

        candidates = []
        if i < starts.size:
            # interval to the right
            d = 0 if (pos >= starts[i] and pos <= ends[i]) else (starts[i] - pos if pos < starts[i] else pos - ends[i])
            candidates.append((i, abs(int(d))))
        if i > 0:
            j = i - 1
            d = 0 if (pos >= starts[j] and pos <= ends[j]) else (starts[j] - pos if pos < starts[j] else pos - ends[j])
            candidates.append((j, abs(int(d))))

        idx, dist = min(candidates, key=lambda t: t[1]) if candidates else (-1, np.iinfo(np.int64).max)
        return idx, dist

    # ---------- public API ----------
    def annotate_gwas(
        self,
        gwas_df: pd.DataFrame,
        window_kb: int = 50,
        chr_col: str = "Chr",
        pos_col: str = "ChrPos",
    ) -> pd.DataFrame:
        """
        Returns a copy of gwas_df with columns:
          - NearestGene
          - NearestGeneID
          - DistanceToGene
          - InWindow  (True if distance <= window_kb*1000)
        """
        if self.genes_df is None or not self.index:
            raise RuntimeError("Load GTF/GFF and call build_index() before annotate_gwas().")

        df = gwas_df.copy()
        if df[chr_col].dtype != object:
            df[chr_col] = df[chr_col].astype(str)
        # normalize 'chr' prefix as index uses normalized strings
        df[chr_col] = df[chr_col].str.replace("^chr", "", regex=True)

        nearest_gene = []
        nearest_id = []
        distance_bp = []
        in_window = []
        max_bp = int(window_kb) * 1000

        for chrom, pos in zip(df[chr_col].tolist(), df[pos_col].tolist()):
            tup = self.index.get(str(chrom))
            if tup is None:
                nearest_gene.append("")
                nearest_id.append("")
                distance_bp.append(np.iinfo(np.int64).max)
                in_window.append(False)
                continue
            starts, ends, names, ids = tup
            idx, dist = self._nearest_distance(int(pos), starts, ends)
            if idx >= 0:
                nearest_gene.append(str(names[idx]))
                nearest_id.append(str(ids[idx]))
                distance_bp.append(int(dist))
                in_window.append(dist <= max_bp)
            else:
                nearest_gene.append("")
                nearest_id.append("")
                distance_bp.append(np.iinfo(np.int64).max)
                in_window.append(False)

        df["NearestGene"] = nearest_gene
        df["NearestGeneID"] = nearest_id
        df["DistanceToGene"] = distance_bp
        df["InWindow"] = in_window
        return df

    def annotate_and_save(
        self,
        gwas_df: pd.DataFrame,
        out_csv: str,
        window_kb: int = 50,
        chr_col: str = "Chr",
        pos_col: str = "ChrPos",
    ) -> str:
        df_annot = self.annotate_gwas(gwas_df, window_kb=window_kb, chr_col=chr_col, pos_col=pos_col)
        out = os.path.splitext(out_csv)[0] + "_annotated.csv"
        df_annot.to_csv(out, index=False)
        return out

    # ---------- optional: region filter ----------
    @staticmethod
    def filter_region(df: pd.DataFrame, region: str, chr_col="Chr", pos_col="ChrPos") -> pd.DataFrame:
        """
        region format: 'chr:start-end' or 'chrom:start-end' (chr prefix optional)
        """
        if not region:
            return df
        region = region.strip()
        if ":" not in region or "-" not in region:
            return df
        chrom, rest = region.split(":", 1)
        start, end = rest.split("-", 1)
        chrom = chrom.replace("chr", "")
        start_i = int(start.replace(",", ""))
        end_i = int(end.replace(",", ""))
        df2 = df.copy()
        if df2[chr_col].dtype != object:
            df2[chr_col] = df2[chr_col].astype(str)
        df2[chr_col] = df2[chr_col].str.replace("^chr", "", regex=True)
        return df2[(df2[chr_col] == chrom) & (df2[pos_col].between(start_i, end_i))]



import os
import dearpygui.dearpygui as dpg

# viewport overlay drawlist
_VP_DL = "__wm_vp_drawlist__"

# main watermark (center)
_TEX = "__wm_texture__"
_IMG = "__wm_image__"

# lab signature (bottom-right)
_LAB_TEX = "__wm_lab_texture__"
_LAB_IMG = "__wm_lab_image__"


def _ensure_viewport_drawlist(front: bool = True):
    if not dpg.does_item_exist(_VP_DL):
        dpg.add_viewport_drawlist(front=front, tag=_VP_DL)
    else:
        dpg.configure_item(_VP_DL, front=front)


def _window_abs_rect(tag: str):
    """Return absolute (pmin, pmax) for `tag` or fallback to viewport size."""
    if not dpg.does_item_exist(tag):
        return (0, 0), (dpg.get_viewport_client_width(), dpg.get_viewport_client_height())

    if hasattr(dpg, "get_item_rect_min") and hasattr(dpg, "get_item_rect_max"):
        try:
            return dpg.get_item_rect_min(tag), dpg.get_item_rect_max(tag)
        except Exception:
            pass

    try:
        pos = dpg.get_item_pos(tag)
        w, h = dpg.get_item_rect_size(tag)
        return pos, (pos[0] + w, pos[1] + h)
    except Exception:
        pass

    return (0, 0), (dpg.get_viewport_client_width(), dpg.get_viewport_client_height())


def _ensure_texture(tag: str, logo_path: str):
    """Create a static texture if not exists; return (width, height) or None."""
    if not os.path.exists(logo_path):
        print(f"[wm] logo not found: {logo_path}")
        return None
    if not dpg.does_item_exist(tag):
        w, h, c, data = dpg.load_image(logo_path)
        with dpg.texture_registry(show=False):
            dpg.add_static_texture(width=w, height=h, default_value=data, tag=tag)
        return w, h
    info = dpg.get_item_configuration(tag)
    return info.get("width", 1), info.get("height", 1)


def _delete_item_if_exists(tag: str):
    if dpg.does_item_exist(tag):
        try:
            dpg.delete_item(tag)
        except Exception:
            pass


def setup(alpha: int = 40,
          scale: float = 0.40,
          target_window_tag: str = "content_area",
          front: bool = True):
    """Draw the main watermark (assets/logo.png) centered inside target window."""
    here = os.path.dirname(__file__)
    logo_path = os.path.abspath(os.path.join(here, "..", "assets", "logo.png"))

    tex_wh = _ensure_texture(_TEX, logo_path)
    if not tex_wh:
        return
    tex_w, tex_h = tex_wh

    _ensure_viewport_drawlist(front=front)

    pmin, pmax = _window_abs_rect(target_window_tag)
    rect_w = max(0, int(pmax[0] - pmin[0]))
    rect_h = max(0, int(pmax[1] - pmin[1]))

    target_w = max(1, int(tex_w * float(scale)))
    target_h = max(1, int(tex_h * float(scale)))

    x = int(pmin[0] + (rect_w - target_w) / 2)
    y = int(pmin[1] + (rect_h - target_h) / 2)

    a = max(0, min(255, int(255 * (float(alpha) / 100.0))))

    _delete_item_if_exists(_IMG)
    dpg.draw_image(
        _TEX,
        pmin=(x, y),
        pmax=(x + target_w, y + target_h),
        uv_min=(0, 0),
        uv_max=(1, 1),
        color=(255, 255, 255, a),
        parent=_VP_DL,
        tag=_IMG,
    )
    print(f"[wm] Watermark placed @ {x},{y} size {target_w}x{target_h} alpha={a}")


def place_signature(target_window_tag: str = "content_area",
                    image_name: str = "logo_lab.png",
                    width: int = 220,
                    margin=(20, 20)):
    """Place the lab logo at bottom-right of the workspace with right-aligned text on its left.
       Repositions existing items if they already exist (responsive)."""
    here = os.path.dirname(__file__)
    img_path = os.path.abspath(os.path.join(here, "..", "assets", image_name))

    tex_wh = _ensure_texture(_LAB_TEX, img_path)
    if not tex_wh or not dpg.does_item_exist(target_window_tag):
        return
    tex_w, tex_h = tex_wh

    # find top-level window to attach items to
    def _top_window(tag: str):
        cur = tag
        for _ in range(20):
            info = dpg.get_item_info(cur)
            if not info:
                break
            if "Window" in info.get("type", ""):
                return cur
            parent = info.get("parent")
            if not parent:
                break
            cur = parent
        return None

    host_window = _top_window(target_window_tag)
    if not host_window:
        return

    # absolute rects
    pmin_ca, pmax_ca = _window_abs_rect(target_window_tag)
    pmin_win, _ = _window_abs_rect(host_window)

    # sizes
    ca_w = int(pmax_ca[0] - pmin_ca[0])
    target_w = int(min(width if width else ca_w * 0.18, tex_w))
    target_h = max(1, int(target_w * (tex_h / max(tex_w, 1))))

    # logo pos (abs -> local)
    abs_logo_x = int(pmax_ca[0] - target_w - int(margin[0]))
    abs_logo_y = int(pmax_ca[1] - target_h - int(margin[1]))
    logo_pos = (abs_logo_x - int(pmin_win[0]), abs_logo_y - int(pmin_win[1]))

    # text (right-aligned block to the left)
    main_text = "Ye-Lab, PKU-IAAS"
    sub_text  = ("Develop by:"
                 " Ahmed Yassin and  Falak Sher Khan")
    gap = 12
    main_w, main_h = dpg.get_text_size(main_text)
    sub_w,  sub_h  = dpg.get_text_size(sub_text)
    text_block_w = max(main_w, sub_w)
    text_block_h = main_h + 4 + sub_h

    text_x = logo_pos[0] - gap - text_block_w
    text_y = int(logo_pos[1] + (target_h - text_block_h) / 2)

    # clamp to keep text inside the content area (left edge)
    ca_local_x0 = int(pmin_ca[0] - pmin_win[0])
    text_x = max(ca_local_x0 + 8, text_x)

    # ensure a container group (so we can move everything together)
    if not dpg.does_item_exist(_LAB_IMG):
        dpg.add_group(parent=host_window, tag=_LAB_IMG)

    # logo item
    _ICON = "__lab_icon__"
    if dpg.does_item_exist(_ICON):
        dpg.configure_item(_ICON, pos=logo_pos, width=target_w, height=target_h, parent=_LAB_IMG)
    else:
        dpg.add_image(_LAB_TEX, pos=logo_pos, width=target_w, height=target_h, parent=_LAB_IMG, tag=_ICON)

    # text items (right-aligned under each other)
    _T1, _T2 = "__lab_t1__", "__lab_t2__"
    if dpg.does_item_exist(_T1):
        dpg.configure_item(_T1, pos=(text_x + (text_block_w - main_w), text_y), parent=_LAB_IMG)
        dpg.configure_item(_T2, pos=(text_x + (text_block_w - sub_w),  text_y + main_h + 4), parent=_LAB_IMG)
    else:
        dpg.add_text(main_text, pos=(text_x + (text_block_w - main_w), text_y),
                     color=(255, 255, 255, 255), parent=_LAB_IMG, tag=_T1)
        dpg.add_text(sub_text,  pos=(text_x + (text_block_w - sub_w),  text_y + main_h + 4),
                     color=(255, 255, 255, 255), parent=_LAB_IMG, tag=_T2)



    print(f"[wm] Logo at {logo_pos} size {target_w}x{target_h}; text at ({text_x},{text_y}) width {text_block_w}")





# plantvarfilter/preanalysis/aligner.py
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, List
import os
import shutil
import subprocess
import time

try:
    from plantvarfilter.variant_caller_utils import resolve_tool  # type: ignore
except Exception:
    resolve_tool = shutil.which  # fallback

try:
    from plantvarfilter.samtools_utils import Samtools  # type: ignore
except Exception:
    Samtools = None  # type: ignore

try:
    from plantvarfilter.preanalysis.reference_manager import ReferenceManager  # type: ignore
except Exception:
    ReferenceManager = None  # type: ignore


@dataclass
class AlignmentResult:
    tool: str
    sam: Optional[str]
    bam: str
    bai: str
    flagstat: str
    elapsed_sec: float
    cmdline: str


class Aligner:
    def __init__(self, logger=print, workspace: Optional[str] = None):
        self.log = logger
        self.workspace = Path(workspace or os.getcwd())
        self.workspace.mkdir(parents=True, exist_ok=True)

        self._paths: Dict[str, Optional[str]] = {
            "samtools": self._tool_path("samtools"),
            "minimap2": self._tool_path("minimap2"),
            "bowtie2": self._tool_path("bowtie2"),
            "bowtie2-build": self._tool_path("bowtie2-build"),
        }

        if Samtools:
            try:
                self.st = Samtools(logger=self.log)  # type: ignore
            except TypeError:
                self.st = Samtools()  # type: ignore
            self._samtools_bin = (
                getattr(self.st, "binary", None)
                or getattr(self.st, "exe", None)
                or self._paths["samtools"]
            )
        else:
            self.st = None
            self._samtools_bin = self._paths["samtools"]

        if not self._samtools_bin:
            raise RuntimeError("samtools not found in PATH")

    def minimap2(
        self,
        reference_mmi_or_fasta: str,
        reads: List[str],
        *,
        preset: str = "map-ont",
        threads: int = 8,
        read_group: Optional[Dict[str, str]] = None,
        save_sam: bool = False,
        mark_duplicates: bool = False,
        out_dir: Optional[str] = None,
        out_prefix: Optional[str] = None,
    ) -> AlignmentResult:
        if not self._paths["minimap2"]:
            raise RuntimeError("minimap2 not found in PATH")

        t0 = time.time()
        out = Path(out_dir or (self.workspace / "alignment")).resolve()
        out.mkdir(parents=True, exist_ok=True)

        ref = Path(reference_mmi_or_fasta)
        if ref.suffix != ".mmi":
            if ReferenceManager is None:
                raise RuntimeError("FASTA provided but .mmi is missing and ReferenceManager is unavailable.")
            rm = ReferenceManager(logger=self.log, workspace=str(self.workspace))  # type: ignore
            status = rm.build_indices(str(ref), out_dir=out.as_posix())
            if not status.mmi:
                raise RuntimeError("Failed to build minimap2 index (.mmi).")
            ref = Path(status.mmi)

        prefix = out_prefix or "aln.minimap2"
        sam_path = out / f"{prefix}.sam"
        bam_path = out / f"{prefix}.sorted.bam"
        bai_path = out / f"{prefix}.sorted.bam.bai"
        flagstat_path = out / f"{prefix}.flagstat.txt"

        rg_args: List[str] = []
        if read_group:
            rg_str = "@RG" + "".join([f"\\t{k}:{v}" for k, v in read_group.items()])
            rg_args = ["-R", rg_str]

        align_cmd = [
            self._paths["minimap2"], "-t", str(threads), "-ax", preset, str(ref), *reads, *rg_args  # type: ignore
        ]
        self.log(f"[ALN] $ {' '.join(align_cmd)}")

        if save_sam:
            self._run(align_cmd + ["-o", str(sam_path)])
            self._sam_to_sorted_bam(str(sam_path), str(bam_path), threads=threads)
        else:
            self._pipe_to_sorted_bam(align_cmd, str(bam_path), threads=threads)

        if mark_duplicates:
            self._markdup_inplace(str(bam_path))

        self._index_bam(str(bam_path))
        self._flagstat(str(bam_path), str(flagstat_path))

        return AlignmentResult(
            tool="minimap2",
            sam=str(sam_path) if save_sam else None,
            bam=str(bam_path),
            bai=str(bai_path),
            flagstat=str(flagstat_path),
            elapsed_sec=time.time() - t0,
            cmdline=" ".join(align_cmd),
        )

    def bowtie2(
        self,
        bt2_prefix: str,
        reads1: str,
        reads2: Optional[str] = None,
        *,
        threads: int = 8,
        read_group: Optional[Dict[str, str]] = None,
        very_sensitive: bool = True,
        save_sam: bool = False,
        mark_duplicates: bool = True,
        out_dir: Optional[str] = None,
        out_prefix: Optional[str] = None,
        extra_args: Optional[List[str]] = None,
    ) -> AlignmentResult:
        if not self._paths["bowtie2"]:
            raise RuntimeError("bowtie2 not found in PATH")

        t0 = time.time()
        out = Path(out_dir or (self.workspace / "alignment")).resolve()
        out.mkdir(parents=True, exist_ok=True)

        prefix = out_prefix or "aln.bowtie2"
        sam_path = out / f"{prefix}.sam"
        bam_path = out / f"{prefix}.sorted.bam"
        bai_path = out / f"{prefix}.sorted.bam.bai"
        flagstat_path = out / f"{prefix}.flagstat.txt"

        args = [self._paths["bowtie2"], "-p", str(threads), "-x", bt2_prefix]  # type: ignore
        if very_sensitive:
            args += ["--very-sensitive"]

        if read_group:
            if "ID" in read_group:
                args += ["--rg-id", str(read_group["ID"])]
            for k in ("SM", "LB", "PL", "PU"):
                if k in read_group:
                    args += ["--rg", f"{k}:{read_group[k]}"]

        if reads2:
            args += ["-1", reads1, "-2", reads2]
        else:
            args += ["-U", reads1]

        if extra_args:
            args += extra_args

        self.log(f"[ALN] $ {' '.join(args)}")

        if save_sam:
            self._run(args + ["-S", str(sam_path)])
            self._sam_to_sorted_bam(str(sam_path), str(bam_path), threads=threads)
        else:
            self._pipe_to_sorted_bam(args + ["-S", "-"], str(bam_path), threads=threads)

        if mark_duplicates:
            self._markdup_inplace(str(bam_path))

        self._index_bam(str(bam_path))
        self._flagstat(str(bam_path), str(flagstat_path))

        return AlignmentResult(
            tool="bowtie2",
            sam=str(sam_path) if save_sam else None,
            bam=str(bam_path),
            bai=str(bai_path),
            flagstat=str(flagstat_path),
            elapsed_sec=time.time() - t0,
            cmdline=" ".join(args),
        )

    def align(
        self,
        platform: str,
        reference: str,
        reads1: str,
        reads2: Optional[str] = None,
        *,
        threads: int = 8,
        read_group: Optional[Dict[str, str]] = None,
        save_sam: bool = False,
        out_dir: Optional[str] = None,
        out_prefix: Optional[str] = None,
    ) -> AlignmentResult:
        p = platform.lower().strip()
        if p in {"ont", "nanopore", "pb", "pacbio", "hifi", "long"}:
            preset = "map-ont" if p in {"ont", "nanopore", "long"} else ("map-hifi" if p in {"hifi"} else "map-pb")
            return self.minimap2(
                reference, [reads1] if not reads2 else [reads1, reads2],
                preset=preset, threads=threads, read_group=read_group,
                save_sam=save_sam, out_dir=out_dir, out_prefix=out_prefix
            )
        return self.bowtie2(
            reference, reads1, reads2,
            threads=threads, read_group=read_group, save_sam=save_sam,
            out_dir=out_dir, out_prefix=out_prefix
        )

    def _tool_path(self, name: str) -> Optional[str]:
        p = resolve_tool(name)
        return str(p) if p else None

    def _run(self, cmd: List[str]) -> None:
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if proc.stdout:
            self.log(proc.stdout.rstrip())
        if proc.returncode != 0:
            raise RuntimeError(f"Command failed: {' '.join(cmd)}")

    def _pipe_to_sorted_bam(self, align_cmd: List[str], bam_out: str, threads: int = 8) -> None:
        assert self._samtools_bin
        p1 = subprocess.Popen(align_cmd, stdout=subprocess.PIPE)
        p2 = subprocess.Popen([self._samtools_bin, "view", "-bS", "-"], stdin=p1.stdout, stdout=subprocess.PIPE)  # type: ignore
        p3 = subprocess.Popen([self._samtools_bin, "sort", "-@", str(threads), "-o", bam_out], stdin=p2.stdout)  # type: ignore
        p3.communicate()
        if p3.returncode != 0:
            raise RuntimeError("Failed to produce sorted BAM")

    def _sam_to_sorted_bam(self, sam_path: str, bam_out: str, threads: int = 8) -> None:
        assert self._samtools_bin
        self._run([self._samtools_bin, "view", "-bS", sam_path, "-o", bam_out])  # type: ignore
        tmp_sorted = bam_out + ".tmp"
        self._run([self._samtools_bin, "sort", "-@", str(threads), "-o", tmp_sorted, bam_out])  # type: ignore
        os.replace(tmp_sorted, bam_out)

    def _markdup_inplace(self, bam_path: str) -> None:
        assert self._samtools_bin
        tmp = bam_path + ".mkdup.tmp.bam"
        try:
            self._run([self._samtools_bin, "markdup", "-r", bam_path, tmp])  # type: ignore
            os.replace(tmp, bam_path)
        except Exception:
            if os.path.exists(tmp):
                try:
                    os.remove(tmp)
                except Exception:
                    pass

    def _index_bam(self, bam_path: str) -> None:
        assert self._samtools_bin
        self._run([self._samtools_bin, "index", bam_path])  # type: ignore

    def _flagstat(self, bam_path: str, out_txt: str) -> None:
        assert self._samtools_bin
        with open(out_txt, "w") as fw:
            proc = subprocess.run([self._samtools_bin, "flagstat", bam_path], stdout=fw, stderr=subprocess.STDOUT, text=True)  # type: ignore
            if proc.returncode != 0:
                raise RuntimeError("samtools flagstat failed")


__all__ = ["Aligner", "AlignmentResult"]



from plantvarfilter.helpers import HELPERS
from sklearn.linear_model import Ridge
from pysnptools.snpreader import Bed, Pheno
import time
from fastlmm.inference import FastLMM
from sklearn.model_selection import KFold
import pandas as pd
from sklearn.model_selection import train_test_split
import xgboost as xgb
from sklearn.ensemble import RandomForestRegressor
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import pearsonr
import seaborn as sns

plt.switch_backend('Agg')


class GenomicPrediction:
    """GenomicPrediction class."""

    def __init__(self):
        #self.gwas_ai = GWASAI()
        self.helper = HELPERS()

    def run_lmm_gp(self, bed_fixed, pheno, genomic_predict_name, model_nr, add_log, bed_file, chrom_mapping):
        """Genomic Prediction using LMM predictor from fast-lmm library."""

        bed_data = Bed(str(bed_file), count_A1=False, chrom_map=chrom_mapping)
        kf = KFold(n_splits=model_nr + 1)
        results = []
        i = 1
        # Start cross-validation
        for train_indices, test_indices in kf.split(range(bed_fixed.iid_count)):
            add_log('Model Iteration: ' + str(i))
            # Split data into training and testing
            train = bed_fixed[train_indices, :]
            # test = bed_fixed[test_indices, :]
            # Train the model
            fastlmm = FastLMM(GB_goal=2)
            fastlmm.fit(K0_train=train, y=pheno)
            # Test the model
            mean, covariance = fastlmm.predict(K0_whole_test=bed_data)
            df = pd.DataFrame({'Column1': mean.val[:, 0], 'Column2': mean.iid[:, 0]})

            # bed_data = bed_fixed.iid
            bed_data2 = bed_data.iid
            pheno_data2 = pheno.iid
            df_gp = pd.DataFrame(bed_data2, columns=['ID1', 'BED_ID2'])
            predicted_values = df['Column1']
            df_gp['Predicted_Value'] = predicted_values

            df_pheno = pd.DataFrame(pheno_data2, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno.read().val
            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(3)

            results.append(merged_df)
            i += 1

        df_all = self.helper.merge_gp_models(results)
        df_all.to_csv(genomic_predict_name, index=False)
        return df_all

    def run_gp_rf(self, bed_fixed, pheno, bed_file, test_size, estimators, genomic_predict_name, chrom_mapping, add_log,
                  model_nr, nr_jobs):
        """Genomic Prediction using Random Forest with cross validation."""
        t1 = time.time()
        dataframes = []
        # df_bim = pd.read_csv(bed_file.replace('bed', 'bim'), delimiter='\t', header=None)
        # df_bim.columns = ['Chr', 'SNP', 'NA', 'ChrPos', 'NA', 'NA']

        # training data
        snp_data = bed_fixed.read().val
        snp_data[np.isnan(snp_data)] = -1
        # entire data for training
        bed_data = Bed(str(bed_file), count_A1=False, chrom_map=chrom_mapping)
        snp_data_all = bed_data.read().val
        snp_data_all[np.isnan(snp_data_all)] = -1
        bed_data2 = bed_data.iid
        pheno_data2 = pheno.iid

        for i in range(int(model_nr)):
            add_log('Model Iteration: ' + str(i + 1))
            # Split data into training and testing sets
            X_train, X_test, y_train, y_test = train_test_split(snp_data, pheno.read().val,
                                                                test_size=test_size)

            # scaler = StandardScaler()
            # X_train = scaler.fit_transform(X_train)
            rf_model = RandomForestRegressor(n_estimators=estimators, n_jobs=nr_jobs)

            rf_model.fit(X_train, y_train.ravel())
            predicted_values = rf_model.predict(snp_data_all)

            df_gp = pd.DataFrame(bed_data2, columns=['ID1', 'BED_ID2'])
            df_gp['Predicted_Value'] = predicted_values

            df_pheno = pd.DataFrame(pheno_data2, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno.read().val
            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(5)
            dataframes.append(merged_df)

        df_all = self.helper.merge_gp_models(dataframes)

        df_all.to_csv(genomic_predict_name, index=False)
        t2 = time.time()
        t3 = round((t2 - t1) / 60, 2)
        add_log('Final run time (minutes): ' + str(t3))
        return df_all

    def run_gp_ridge(self, bed_fixed, pheno, bed_file, test_size, alpha, genomic_predict_name, chrom_mapping, add_log,
                     model_nr):
        """Genomic Prediction using Ridge Regression with cross validation."""

        t1 = time.time()
        dataframes = []
        # df_bim = pd.read_csv(bed_file.replace('bed', 'bim'), delimiter='\t', header=None)
        # df_bim.columns = ['Chr', 'SNP', 'NA', 'ChrPos', 'NA', 'NA']

        # training data
        snp_data = bed_fixed.read().val
        snp_data[np.isnan(snp_data)] = -1
        # entire data for training
        bed_data = Bed(str(bed_file), count_A1=False, chrom_map=chrom_mapping)
        snp_data_all = bed_data.read().val
        snp_data_all[np.isnan(snp_data_all)] = -1
        bed_data2 = bed_data.iid
        pheno_data2 = pheno.iid

        for i in range(int(model_nr)):
            add_log('Model Iteration: ' + str(i + 1))
            # Split data into training and testing sets
            X_train, X_test, y_train, y_test = train_test_split(snp_data, pheno.read().val,
                                                                test_size=test_size)

            # scaler = StandardScaler()
            # X_train = scaler.fit_transform(X_train)
            ridge_model = Ridge(alpha=alpha)

            ridge_model.fit(X_train, y_train.ravel())
            predicted_values = ridge_model.predict(snp_data_all)

            df_gp = pd.DataFrame(bed_data2, columns=['ID1', 'BED_ID2'])
            df_gp['Predicted_Value'] = predicted_values

            df_pheno = pd.DataFrame(pheno_data2, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno.read().val
            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(5)
            dataframes.append(merged_df)

        df_all = self.helper.merge_gp_models(dataframes)

        df_all.to_csv(genomic_predict_name, index=False)
        t2 = time.time()
        t3 = round((t2 - t1) / 60, 2)
        add_log('Final run time (minutes): ' + str(t3))
        return df_all

    def run_gp_xg(self, bed_fixed, pheno, bed_file, test_size, estimators, genomic_predict_name, chrom_mapping, add_log,
                  model_nr, max_dep_set, nr_jobs):
        """Genomic Prediction using XgBOOST with cross validation."""

        t1 = time.time()
        dataframes = []
        # df_bim = pd.read_csv(bed_file.replace('bed', 'bim'), delimiter='\t', header=None)
        # df_bim.columns = ['Chr', 'SNP', 'NA', 'ChrPos', 'NA', 'NA']
        # print (len(pheno.read().val))
        validation = False
        if validation:

            snp_data = bed_fixed.read().val
            pheno_data = pheno.read().val

            # Assuming x and y are your numpy arrays and they have the same length
            total_data_points = len(pheno_data)

            # Calculate the number of validation samples (5% of the total data)
            num_validation_samples = round(0.1 * total_data_points)

            # Generate random indices for the validation set
            indices = np.arange(total_data_points)
            np.random.shuffle(indices)
            val_indices = indices[:num_validation_samples]
            train_test_indices = indices[num_validation_samples:]

            # Create validation set
            snps_for_validation = snp_data[val_indices]
            pheno_for_validation = pheno_data[val_indices]
            ids_for_validation1 = bed_fixed.iid[val_indices]
            ids_for_validation2 = pheno.iid[val_indices]
            # print (ids_for_validation1)
            # print(ids_for_validation2, pheno_for_validation)

            # Remaining data for training/testing
            snps_for_training = snp_data[train_test_indices]
            pheno_for_training = pheno_data[train_test_indices]
            # print (bed_fixed.iid[train_test_indices])
            # np.savetxt('rest.csv', pheno_for_training, delimiter=',')
            # np.savetxt('val.csv', pheno_for_validation, delimiter=',')
        else:

            snps_for_training = bed_fixed.read().val
            pheno_for_training = pheno.read().val

        # print (len(snps_for_training), len(pheno_for_training))

        for i in range(int(model_nr)):
            add_log('Model Iteration: ' + str(i + 1))

            # Split data into training and testing sets
            # X_train, X_test, y_train, y_test = train_test_split(bed_fixed.read().val, pheno.read().val,
            # test_size=test_size)
            X_train, X_test, y_train, y_test = train_test_split(snps_for_training, pheno_for_training,
                                                                test_size=test_size)

            # scaler = StandardScaler()
            # X_train = scaler.fit_transform(X_train)
            xgb_model = xgb.XGBRegressor(n_estimators=estimators, learning_rate=0.1,
                                         max_depth=max_dep_set, nthread=nr_jobs)  # , min_child_weight=3)
            xgb_model.fit(X_train, y_train)

            # for prediction, we need all genotypes
            bed_data = Bed(str(bed_file), count_A1=False, chrom_map=chrom_mapping)
            bed_data2 = bed_data.iid

            pheno_data2 = pheno.iid
            predicted_values = xgb_model.predict(bed_data.read().val)

            df_gp = pd.DataFrame(bed_data2, columns=['ID1', 'BED_ID2'])
            df_gp['Predicted_Value'] = predicted_values

            df_pheno = pd.DataFrame(pheno_data2, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno.read().val
            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(5)
            dataframes.append(merged_df)

        df_all = self.helper.merge_gp_models(dataframes)
        df_all.to_csv(genomic_predict_name, index=False)

        if validation:
            unique_ids = np.unique(ids_for_validation1.flatten())
            # Filter the DataFrame based on these IDs
            valdiation_df = df_all[df_all['ID1'].isin(unique_ids)]

            valdiation_df.to_csv(genomic_predict_name.replace('.csv', '_valdation.csv'), index=False)
            # print(filtered_df)
        else:
            valdiation_df = None

        t2 = time.time()
        t3 = round((t2 - t1) / 60, 2)
        add_log('Final run time (minutes): ' + str(t3))
        return df_all
        #return df_all, valdiation_df

    def find_best_parameters(self, snps_for_training, pheno_for_training, param_grid):
        from sklearn.model_selection import GridSearchCV, KFold, train_test_split
        xgb_model = xgb.XGBRegressor(objective='reg:squarederror')
        cv = KFold(n_splits=2, shuffle=True, random_state=42)

        grid_search = GridSearchCV(estimator=xgb_model, param_grid=param_grid, cv=cv, scoring='neg_mean_squared_error',
                                   n_jobs=-1)
        grid_search.fit(snps_for_training, pheno_for_training)

        best_params = grid_search.best_params_
        print(f"Best parameters found: {best_params}")
        # Print the mean and standard deviation of the test scores for each parameter combination
        means = grid_search.cv_results_['mean_test_score']
        stds = grid_search.cv_results_['std_test_score']
        params = grid_search.cv_results_['params']

        for mean, std, param in zip(means, stds, params):
            print(f"Params: {param}, Mean MSE: {-mean:.4f}, Std MSE: {std:.4f}")

        return best_params

    def model_validation(self, bed_fixed, pheno, bed_file, test_size, estimators, genomic_predict_name, chrom_mapping, add_log,
                  model_nr, max_dep_set, validation_size=0.1):
        # Define your parameter grid
        param_grid = {
            'n_estimators': [100, 200, 300],
            'learning_rate': [0.01, 0.1, 0.2],
            'max_depth': [3, 5, 7]
        }
        #
        param_grid = {
            'n_estimators': [100],
            'learning_rate': [0.01],
            'max_depth': [3],
            'alpha': [0],  # L1 regularization
            'lambda': [0],  # L2 regularization
           # 'scale_pos_weight': [4]
        }

        snps_for_training = bed_fixed.read().val
        #snps_for_training[np.isnan(snps_for_training)] = -1

        pheno_for_training = pheno.read().val
        ids = pheno.iid

        # Find the best parameters
        best_params = self.find_best_parameters(snps_for_training, pheno_for_training, param_grid)

        # Create validation set (10% of the data)
        X_train_full, X_val, y_train_full, y_val, train_full_ids, val_ids = train_test_split(
            snps_for_training, pheno_for_training, ids, test_size=0.2)#, stratify=pheno_for_training)

        #from sklearn.preprocessing import StandardScaler
        #scaler = StandardScaler()
       # X_train_full = scaler.fit_transform(X_train_full)
        #y_train_full = scaler.fit_transform(y_train_full)


        # Initialize results storage
        all_models_predictions = []

        # Perform 10 iterations of training/testing excluding the validation set
        for i in range(50):
        # kf = KFold(n_splits=2, shuffle=True)
        # for train_index, test_index in kf.split(X_train_full):
        #     X_train, X_test = X_train_full[train_index], X_train_full[test_index]
        #     y_train, y_test = y_train_full[train_index], y_train_full[test_index]
        #     train_ids, test_ids = train_full_ids[train_index], train_full_ids[test_index]

            # # Further split the training data into training and testing sets (70/30 split)
            X_train, X_test, y_train, y_test, train_ids, test_ids = train_test_split(
                X_train_full, y_train_full, train_full_ids, test_size=test_size)#, stratify=y_train_full)


            # Initialize and train the model with the best parameters
            xgb_model = xgb.XGBRegressor(**best_params)
            xgb_model.fit(X_train, y_train)

            # Predictions for the training, test, and validation sets
            train_preds = xgb_model.predict(X_train)
            test_preds = xgb_model.predict(X_test)
            val_preds = xgb_model.predict(X_val)

            # Store predictions for analysis
            model_results = {
                'ID': np.concatenate([train_ids, test_ids, val_ids]),
                'Phenotype': np.concatenate([y_train, y_test, y_val]),
                'Prediction': np.concatenate([train_preds, test_preds, val_preds]),
                'Set': ['Training'] * len(train_preds) + ['Testing'] * len(test_preds) + ['Validation'] * len(val_preds)
            }
            model_results['ID'] = model_results['ID'][:, 0]
            if model_results['Phenotype'].ndim > 1:
                model_results['Phenotype'] = model_results['Phenotype'].ravel()

            all_models_predictions.append(model_results)


            # Optionally create a DataFrame and save to CSV for each iteration if needed
            df = pd.DataFrame(model_results)
            df.to_csv(f'predictions_model_{len(all_models_predictions)}.csv', index=False)


        df_all = pd.concat([pd.DataFrame(model) for model in all_models_predictions], ignore_index=True)

        # Group by 'ID' and calculate mean and SD of 'Prediction'
        summary_stats = df_all.groupby('ID')['Prediction'].agg(['mean', 'std']).reset_index()
        summary_stats.columns = ['ID', 'Mean_Prediction', 'SD_Prediction']
        # Dropping duplicates since we assume 'Phenotype' and 'Set' do not change per ID
        original_data = df_all[['ID', 'Phenotype', 'Set']].drop_duplicates()
        # Merge summary statistics with original data
        final_data = pd.merge(original_data, summary_stats, on='ID', how='left')
        final_data.to_csv('final_predictions_summary.csv', index=False)

        plt.figure(figsize=(18, 6))

        # Create a subplot for each set
        for i, set_type in enumerate(['Training', 'Testing', 'Validation'], 1):
            plt.subplot(1, 3, i)
            subset = final_data[final_data['Set'] == set_type]

            # Plotting the correlation
            correlation = subset['Phenotype'].corr(subset['Mean_Prediction'])
            sns.scatterplot(x='Phenotype', y='Mean_Prediction', data=subset)
            sns.regplot(x='Phenotype', y='Mean_Prediction', data=subset, scatter=False, color='red')

            # Title with correlation coefficient
            plt.title(f'{set_type} Set (corr={correlation:.2f})')
            plt.xlabel('Phenotype')
            plt.ylabel('Mean_Prediction')

        # Adjust layout
        plt.tight_layout()

        # Save the plot to a file
        plt.savefig('correlation_plots.png', dpi=300)  # Save as high-resolution PNG file

        # Show plot
        #plt.show()


    def plot_gp(self, df, gp_plot_name, algorithm):
        """Bland-Altman Plot for the real and predicted phenotype values."""
        #print (df)
        # Calculate mean and difference (redundant here, but for demonstration)
        df['Mean'] = (df['Pheno_Value'] + df['Mean_Predicted_Value']) / 2
        df['Difference'] = df['Pheno_Value'] - df['Mean_Predicted_Value']

        # Plotting the Bland-Altman Plot
        plt.figure(figsize=(10, 6))
        plt.scatter(df['Mean'], df['Difference'], color='blue')
        # sns.scatterplot(x='Mean', y='Difference', data=df, color='blue')

        # Calculate and plot the mean difference
        mean_diff = df['Difference'].mean()
        plt.axhline(mean_diff, color='red', linestyle='--')

        # Calculate and plot the limits of agreement
        std_diff = df['Difference'].std()

        plt.axhline(mean_diff, color='red', linestyle='--', label='Mean Difference')
        plt.axhline(mean_diff + 1.96 * std_diff, color='green', linestyle='--', label='Upper Limit of Agreement')
        plt.axhline(mean_diff - 1.96 * std_diff, color='green', linestyle='--', label='Lower Limit of Agreement')
        plt.xlabel('Mean Value')
        plt.ylabel('Difference')
        algorithm2 = algorithm.replace(' (AI)', '')
        plt.title('Bland-Altman Plot (' + algorithm2 + ')')
        plt.legend()

        plt.tight_layout(pad=1)
        plt.savefig(gp_plot_name)
        plt.savefig(gp_plot_name.replace('Bland_Altman_plot', 'Bland_Altman_plot_high'), dpi=300)

    def plot_gp_scatter(self, df, gp_plot_name_scatter, algorithm):
        """Regression Plot for the real and predicted phenotype values."""

        df = df.replace([np.inf, -np.inf], np.nan).dropna()

        # Pearson correlation
        corr, _ = pearsonr(df['Pheno_Value'], df['Mean_Predicted_Value'])
        corr_label = f"Pearson correlation: {corr:.2f}"

        # Plotting
        plt.figure(figsize=(10, 6))
        sns.scatterplot(data=df, x='Pheno_Value', y='Mean_Predicted_Value')
        sns.regplot(x='Pheno_Value', y='Mean_Predicted_Value', data=df, scatter=False, color='red')
        plt.title('Scatter Plot with Regression Line')
        plt.xlabel('Phenotype Value')
        plt.ylabel('Mean Predicted Value')
        # plt.text(5, max(df['Mean_Predicted_Value']) - 5, corr_label, fontsize=12, color='blue')
        algorithm2 = algorithm.replace(' (AI)', '')
        plt.title('Correlation Plot (' + algorithm2 + ')' + '\n' + corr_label)
        plt.tight_layout(pad=1)
        plt.savefig(gp_plot_name_scatter)
        plt.savefig(gp_plot_name_scatter.replace('GP_scatter_plot', 'GP_scatter_plot_high'), dpi=300)
        # plt.show()



# ld_utils.py
import os
import re
import math
import csv
import subprocess
from typing import Optional, Tuple, List, Dict

try:
    from plantvarfilter.linux import resolve_tool
except Exception:
    from shutil import which as _which
    def resolve_tool(name: str) -> str:
        return _which(name) or name


class LDAnalysisError(Exception):
    pass


def _run(cmd: List[str]) -> str:
    p = subprocess.run(cmd, text=True, capture_output=True)
    if p.returncode != 0:
        raise LDAnalysisError(f"Command failed ({p.returncode}): {' '.join(cmd)}\n{p.stdout}\n{p.stderr}")
    return p.stdout or ""


def _parse_region(region: Optional[str]) -> Tuple[Optional[str], Optional[int], Optional[int]]:
    if not region or not str(region).strip():
        return None, None, None
    m = re.match(r"^(\w+):(\d+)[-:](\d+)$", region.replace(",", ""))
    if not m:
        raise LDAnalysisError("Region must look like: chr1:100000-200000")
    chrom, start, end = m.group(1), int(m.group(2)), int(m.group(3))
    if end < start:
        start, end = end, start
    return chrom, start, end


def _ensure_dir(path: str):
    os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)


class LDAnalyzer:
    def __init__(self, plink_bin: Optional[str] = None):
        self.plink = plink_bin or resolve_tool("plink")

    def _plink_base(
        self,
        bfile_prefix: Optional[str] = None,
        vcf_path: Optional[str] = None,
        out_prefix: Optional[str] = None,
        keep_samples: Optional[str] = None,
        region: Optional[str] = None,
    ) -> List[str]:
        if not out_prefix:
            raise LDAnalysisError("out_prefix is required")
        _ensure_dir(out_prefix + ".x")
        cmd = [self.plink, "--allow-no-sex", "--silent", "--out", out_prefix]
        if bfile_prefix:
            cmd += ["--bfile", bfile_prefix]
        elif vcf_path:
            cmd += ["--vcf", vcf_path]
        else:
            raise LDAnalysisError("Either bfile_prefix or vcf_path must be provided")
        if keep_samples:
            cmd += ["--keep", keep_samples]
        if region:
            chrom, start, end = _parse_region(region)
            if chrom:
                cmd += ["--chr", str(chrom), "--from-bp", str(start), "--to-bp", str(end)]
        return cmd

    # ---------- LD decay ----------
    def ld_decay(
        self,
        out_prefix: str,
        bfile_prefix: Optional[str] = None,
        vcf_path: Optional[str] = None,
        window_kb: int = 500,
        window_snps: int = 5000,
        max_dist_kb: int = 1000,
        min_r2: float = 0.1,
        region: Optional[str] = None,
        keep_samples: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Produces:
          - <out_prefix>.ld.gz        (pairwise LD list)
          - <out_prefix>.decay.csv    (binned mean r^2 vs distance)
          - <out_prefix>.decay.png    (plot)
        """
        # 1) pairwise LD with distances (bp)
        cmd = self._plink_base(bfile_prefix, vcf_path, out_prefix, keep_samples, region)
        cmd += [
            "--r2", "yes-really", "gz",
            "--ld-window-kb", str(window_kb),
            "--ld-window", str(window_snps),
            "--ld-window-r2", str(min_r2),
            "--ld-window-bp", str(max_dist_kb * 1000),
        ]
        _run(cmd)

        # 2) Summarize decay curve from .ld.gz
        ld_gz = out_prefix + ".ld.gz"
        if not os.path.exists(ld_gz):
            raise LDAnalysisError("plink did not create .ld.gz")

        # Use zcat or python gzip to read; prefer python for portability
        import gzip
        dists, r2s = [], []
        with gzip.open(ld_gz, "rt") as fh:
            header = fh.readline().strip().split()
            # Expect columns: CHR_A BP_A SNP_A CHR_B BP_B SNP_B R2
            # plink v1.9 has: CHR_A BP_A SNP_A CHR_B BP_B SNP_B R^2
            # normalize key:
            r2_key = "R2" if "R2" in header else "R^2"
            idx_bp_a = header.index("BP_A")
            idx_bp_b = header.index("BP_B")
            idx_r2 = header.index(r2_key)
            for ln in fh:
                parts = ln.strip().split()
                try:
                    d = abs(int(parts[idx_bp_b]) - int(parts[idx_bp_a]))
                    r = float(parts[idx_r2])
                except Exception:
                    continue
                dists.append(d)
                r2s.append(r)

        if not dists:
            raise LDAnalysisError("No LD pairs returned by plink (check filters/region)")

        # Bin by distance (kb) and compute mean r^2
        max_bp = max(dists)
        bin_size_bp = max(1, int(max(1000, (max_bp // 50))))  # ~50 bins
        bins: Dict[int, List[float]] = {}
        for d, r in zip(dists, r2s):
            b = (d // bin_size_bp) * bin_size_bp
            bins.setdefault(b, []).append(r)

        rows = []
        for b in sorted(bins.keys()):
            arr = bins[b]
            mean_r2 = sum(arr) / len(arr)
            kb = b / 1000.0
            rows.append((kb, mean_r2))

        decay_csv = out_prefix + ".decay.csv"
        with open(decay_csv, "w", newline="") as f:
            w = csv.writer(f)
            w.writerow(["distance_kb", "mean_r2"])
            for kb, r in rows:
                w.writerow([f"{kb:.6f}", f"{r:.6f}"])

        # 3) Plot
        try:
            import matplotlib
            matplotlib.use("Agg")
            import matplotlib.pyplot as plt
            xs = [r[0] for r in rows]
            ys = [r[1] for r in rows]
            plt.figure(figsize=(6, 4))
            plt.plot(xs, ys)
            plt.xlabel("Distance (kb)")
            plt.ylabel("Mean r²")
            plt.title("LD decay")
            plt.tight_layout()
            decay_png = out_prefix + ".decay.png"
            plt.savefig(decay_png, dpi=150)
            plt.close()
        except Exception:
            decay_png = ""

        return {"pairs_gz": ld_gz, "decay_csv": decay_csv, "decay_png": decay_png}

    # ---------- LD heatmap ----------
    def ld_heatmap(
        self,
        out_prefix: str,
        bfile_prefix: Optional[str] = None,
        vcf_path: Optional[str] = None,
        region: Optional[str] = None,
        window_snps: int = 1000,
        min_r2: float = 0.0,
        keep_samples: Optional[str] = None,
        max_snps_for_plot: int = 500,
    ) -> Dict[str, str]:
        """
        Produces a square r^2 matrix heatmap for the given region.
        If region is None, plink will use all SNPs up to window_snps.
        """
        # 1) get pairwise LD list
        cmd = self._plink_base(bfile_prefix, vcf_path, out_prefix, keep_samples, region)
        cmd += ["--r2", "yes-really", "gz", "--ld-window", str(window_snps), "--ld-window-r2", str(min_r2)]
        _run(cmd)
        ld_gz = out_prefix + ".ld.gz"

        # 2) build matrix from list
        import gzip
        snp_index: Dict[str, int] = {}
        pairs: List[Tuple[int, int, float]] = []

        with gzip.open(ld_gz, "rt") as fh:
            header = fh.readline().strip().split()
            r2_key = "R2" if "R2" in header else "R^2"
            i_snp_a = header.index("SNP_A")
            i_snp_b = header.index("SNP_B")
            i_r2 = header.index(r2_key)
            for ln in fh:
                p = ln.strip().split()
                a, b = p[i_snp_a], p[i_snp_b]
                r2 = float(p[i_r2])
                if a not in snp_index:
                    snp_index[a] = len(snp_index)
                if b not in snp_index:
                    snp_index[b] = len(snp_index)
                pairs.append((snp_index[a], snp_index[b], r2))

        n = len(snp_index)
        # downsample if too big for plotting
        if n > max_snps_for_plot:
            # keep first N by their order of appearance
            keep = set(i for i, _ in zip(range(n), range(max_snps_for_plot)))
            pairs = [(i, j, r) for (i, j, r) in pairs if i in keep and j in keep]
            n = max_snps_for_plot

        import numpy as np
        mat = np.ones((n, n), dtype=float)
        mat[:] = np.nan
        for i, j, r in pairs:
            if i < n and j < n:
                mat[i, j] = r
                mat[j, i] = r
        for i in range(n):
            mat[i, i] = 1.0

        # 3) plot
        try:
            import matplotlib
            matplotlib.use("Agg")
            import matplotlib.pyplot as plt
            plt.figure(figsize=(6, 5))
            im = plt.imshow(mat, interpolation="nearest", origin="lower", vmin=0.0, vmax=1.0)
            plt.colorbar(im, fraction=0.046, pad=0.04, label="r²")
            plt.title("LD heatmap")
            plt.tight_layout()
            heat_png = out_prefix + ".heatmap.png"
            plt.savefig(heat_png, dpi=150)
            plt.close()
        except Exception:
            heat_png = ""

        return {"pairs_gz": ld_gz, "heat_png": heat_png}

    # ---------- Diversity metrics ----------
    def diversity(
        self,
        out_prefix: str,
        bfile_prefix: Optional[str] = None,
        vcf_path: Optional[str] = None,
        keep_samples: Optional[str] = None,
        region: Optional[str] = None,
    ) -> Dict[str, str]:
        """
        Uses plink to compute common per-variant and per-sample metrics:
          --freq  (MAF)
          --hardy (HWE)
          --het   (observed/expected het per individual)
        Produces:
          <out_prefix>.frq, <out_prefix>.hwe, <out_prefix>.het and a summary CSV.
        """
        base = self._plink_base(bfile_prefix, vcf_path, out_prefix, keep_samples, region)
        # freq
        _run(base + ["--freq"])
        # hardy
        _run(base + ["--hardy"])
        # sample het
        _run(base + ["--het"])

        # Summarize
        frq, hwe, het = out_prefix + ".frq", out_prefix + ".hwe", out_prefix + ".het"
        summary_csv = out_prefix + ".diversity.csv"

        # quick aggregation: mean MAF, fraction of SNPs with MAF<0.05, HWE fails, mean F
        try:
            import pandas as pd
            df_frq = pd.read_csv(frq, delim_whitespace=True) if os.path.exists(frq) else None
            df_hwe = pd.read_csv(hwe, delim_whitespace=True) if os.path.exists(hwe) else None
            df_het = pd.read_csv(het, delim_whitespace=True) if os.path.exists(het) else None

            rows = []
            if df_frq is not None and "MAF" in df_frq.columns:
                mean_maf = float(df_frq["MAF"].replace(".", "0").astype(float).mean())
                low05 = float((df_frq["MAF"].replace(".", "0").astype(float) < 0.05).mean())
                rows.append(("mean_maf", mean_maf))
                rows.append(("fraction_maf_lt_0.05", low05))
            if df_hwe is not None and "P" in df_hwe.columns:
                fails = float((df_hwe["P"].replace(".", "1").astype(float) < 1e-6).mean())
                rows.append(("hwe_fail_fraction_p_lt_1e-6", fails))
            if df_het is not None and "F" in df_het.columns:
                mean_F = float(df_het["F"].astype(float).mean())
                rows.append(("mean_inbreeding_coeff_F", mean_F))

            with open(summary_csv, "w", newline="") as f:
                w = csv.writer(f)
                w.writerow(["metric", "value"])
                for k, v in rows:
                    w.writerow([k, f"{v:.6g}"])
        except Exception:
            # fallback: just create an empty summary
            with open(summary_csv, "w", newline="") as f:
                csv.writer(f).writerow(["metric", "value"])

        return {"frq": frq, "hwe": hwe, "het": het, "summary_csv": summary_csv}



# ui/ui_theme.py
import os
import dearpygui.dearpygui as dpg
from contextlib import contextmanager

# ---------------- Design tokens ----------------
FONT_SCALE = 1.10
SPACING = 14
RADIUS = 10
CARD_W, CARD_H = 560, 280

# ---------------- Palette ----------------
COLOR_TEXT_DARK = (234, 238, 241)
COLOR_TEXT_LIGHT = (28, 32, 33)

COLOR_BG_DARK_WINDOW = (22, 25, 26)
COLOR_BG_DARK_CHILD  = (30, 34, 35)
COLOR_BG_DARK_FRAME  = (42, 46, 47)

COLOR_BG_LIGHT_WINDOW = (246, 247, 248)
COLOR_BG_LIGHT_CHILD  = (236, 239, 242)
COLOR_BG_LIGHT_FRAME  = (225, 228, 232)

COLOR_TITLE  = (205, 228, 210)
COLOR_NOTE   = (219, 191, 84)
COLOR_MUTED  = (152, 163, 169)

# Default primary (will be overridden at runtime)
COLOR_PRIMARY   = (46, 125, 50)
COLOR_PRIMARY_H = (67, 160, 71)
COLOR_PRIMARY_A = (27, 94, 32)

# ---------------- Font helpers ----------------
_APP_FONT_TAG = "ui_font_base"
_FONT_REGISTRY_TAG = "font_registry"

def _project_fonts_dir() -> str:
    here = os.path.dirname(__file__)
    return os.path.abspath(os.path.join(here, "..", "assets", "fonts"))

def _resolve_font_path(font_dir: str | None, font_file: str | None) -> str | None:
    if font_file:
        if os.path.isabs(font_file) and os.path.exists(font_file):
            return font_file
        if font_dir:
            p = os.path.join(font_dir, font_file)
            if os.path.exists(p):
                return p

    default_project_font = os.path.join(_project_fonts_dir(), "test.ttf")
    if os.path.exists(default_project_font):
        return default_project_font

    here = os.path.dirname(__file__)
    candidates = [p for p in [
        font_dir,
        os.path.join(here, "assets", "fonts"),
        _project_fonts_dir()
    ] if p]

    common_files = [
        "test.ttf",
        "Inter.ttf", "Inter-Regular.ttf",
        "NotoSans-Regular.ttf",
        "DejaVuSans.ttf"
    ]
    for root in candidates:
        if not os.path.isdir(root):
            continue
        for fn in common_files:
            p = os.path.join(root, fn)
            if os.path.exists(p):
                return p
    return None

def _register_font(font_path: str, base_size: int = 18):
    if dpg.does_item_exist(_FONT_REGISTRY_TAG):
        dpg.delete_item(_FONT_REGISTRY_TAG)
    with dpg.font_registry(tag=_FONT_REGISTRY_TAG):
        with dpg.font(font_path, base_size, tag=_APP_FONT_TAG):
            for hint_name in (
                "mvFontRangeHint_Default",
                "mvFontRangeHint_Cyrillic",
                "mvFontRangeHint_Korean",
                "mvFontRangeHint_Japanese",
                "mvFontRangeHint_Thai",
                "mvFontRangeHint_Vietnamese",
                "mvFontRangeHint_Chinese_Full",
            ):
                if hasattr(dpg, hint_name):
                    dpg.add_font_range_hint(getattr(dpg, hint_name))
    dpg.bind_font(_APP_FONT_TAG)
    print(f"[THEME] Loaded font: {font_path} @ {base_size}px")

# ---------------- App chrome & global ----------------
def setup_app_chrome(
    font_scale: float | None = None,
    font_dir: str | None = None,
    base_size: int = 18,
    font_file: str | None = None,
):
    dpg.configure_app(docking=True, docking_space=True)
    dpg.set_global_font_scale(font_scale if font_scale else FONT_SCALE)

    if dpg.does_item_exist(_APP_FONT_TAG):
        try:
            dpg.bind_font(_APP_FONT_TAG)
            return
        except Exception:
            pass

    if font_dir is None and font_file is None:
        font_dir = _project_fonts_dir()
        font_file = "test.ttf"

    path = _resolve_font_path(font_dir=font_dir, font_file=font_file)
    if path:
        try:
            _register_font(path, base_size=base_size)
        except Exception as e:
            print(f"[THEME] Font load failed: {e}")
    else:
        print("[THEME] No custom font found. Using Dear ImGui default.")

def set_font_scale(scale: float):
    try:
        dpg.set_global_font_scale(float(scale))
    except Exception:
        pass

# ---------------- Global themes (static) ----------------
def build_dark_theme():
    if dpg.does_item_exist("theme_dark"):
        return
    with dpg.theme(tag="theme_dark"):
        with dpg.theme_component(dpg.mvAll):
            dpg.add_theme_style(dpg.mvStyleVar_WindowPadding, SPACING, SPACING)
            dpg.add_theme_style(dpg.mvStyleVar_ItemSpacing, SPACING, SPACING)
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 9, 7)
            dpg.add_theme_style(dpg.mvStyleVar_WindowRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_ChildRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_PopupRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_GrabRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_TabRounding, RADIUS)

            dpg.add_theme_color(dpg.mvThemeCol_Text, COLOR_TEXT_DARK)
            dpg.add_theme_color(dpg.mvThemeCol_TextDisabled, (160, 165, 170))
            dpg.add_theme_color(dpg.mvThemeCol_WindowBg, COLOR_BG_DARK_WINDOW)
            dpg.add_theme_color(dpg.mvThemeCol_ChildBg, COLOR_BG_DARK_CHILD)
            dpg.add_theme_color(dpg.mvThemeCol_PopupBg, COLOR_BG_DARK_CHILD)
            dpg.add_theme_color(dpg.mvThemeCol_FrameBg, COLOR_BG_DARK_FRAME)
            dpg.add_theme_color(dpg.mvThemeCol_FrameBgHovered, (58, 62, 64))
            dpg.add_theme_color(dpg.mvThemeCol_FrameBgActive, (66, 70, 72))
            dpg.add_theme_color(dpg.mvThemeCol_Border, (70, 74, 78))
            dpg.add_theme_color(dpg.mvThemeCol_Separator, (70, 74, 78))

            dpg.add_theme_color(dpg.mvThemeCol_TitleBg, (18, 48, 42))
            dpg.add_theme_color(dpg.mvThemeCol_TitleBgActive, (24, 72, 63))
            dpg.add_theme_color(dpg.mvThemeCol_Tab, (28, 52, 47))
            dpg.add_theme_color(dpg.mvThemeCol_TabHovered, (34, 66, 60))
            dpg.add_theme_color(dpg.mvThemeCol_TabActive, (26, 58, 52))

def build_light_theme():
    if dpg.does_item_exist("theme_light"):
        return
    with dpg.theme(tag="theme_light"):
        with dpg.theme_component(dpg.mvAll):
            dpg.add_theme_style(dpg.mvStyleVar_WindowPadding, SPACING, SPACING)
            dpg.add_theme_style(dpg.mvStyleVar_ItemSpacing, SPACING, SPACING)
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 9, 7)
            dpg.add_theme_style(dpg.mvStyleVar_WindowRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_ChildRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_PopupRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_GrabRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_TabRounding, RADIUS)

            dpg.add_theme_color(dpg.mvThemeCol_Text, COLOR_TEXT_LIGHT)
            dpg.add_theme_color(dpg.mvThemeCol_TextDisabled, (110, 120, 128))
            dpg.add_theme_color(dpg.mvThemeCol_WindowBg, COLOR_BG_LIGHT_WINDOW)
            dpg.add_theme_color(dpg.mvThemeCol_ChildBg, COLOR_BG_LIGHT_CHILD)
            dpg.add_theme_color(dpg.mvThemeCol_PopupBg, COLOR_BG_LIGHT_CHILD)
            dpg.add_theme_color(dpg.mvThemeCol_FrameBg, COLOR_BG_LIGHT_FRAME)
            dpg.add_theme_color(dpg.mvThemeCol_FrameBgHovered, (218, 222, 226))
            dpg.add_theme_color(dpg.mvThemeCol_FrameBgActive, (210, 214, 219))
            dpg.add_theme_color(dpg.mvThemeCol_Border, (205, 210, 215))
            dpg.add_theme_color(dpg.mvThemeCol_Separator, (205, 210, 215))

            dpg.add_theme_color(dpg.mvThemeCol_TitleBg, (210, 232, 224))
            dpg.add_theme_color(dpg.mvThemeCol_TitleBgActive, (194, 220, 212))
            dpg.add_theme_color(dpg.mvThemeCol_Tab, (220, 234, 228))
            dpg.add_theme_color(dpg.mvThemeCol_TabHovered, (208, 226, 220))
            dpg.add_theme_color(dpg.mvThemeCol_TabActive, (200, 220, 214))

def apply_theme(dark: bool = True):
    dpg.bind_theme("theme_dark" if dark else "theme_light")

# ---------------- Component themes (static baseline) ----------------
def build_component_themes():
    if not dpg.does_item_exist("theme_button_primary"):
        with dpg.theme(tag="theme_button_primary"):
            with dpg.theme_component(dpg.mvButton):
                dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
                dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 12, 8)
                dpg.add_theme_color(dpg.mvThemeCol_Button,        (*COLOR_PRIMARY, 255))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, (*COLOR_PRIMARY_H, 255))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonActive,  (*COLOR_PRIMARY_A, 255))
                dpg.add_theme_color(dpg.mvThemeCol_Text,          (255, 255, 255))

    if not dpg.does_item_exist("theme_button_secondary"):
        with dpg.theme(tag="theme_button_secondary"):
            with dpg.theme_component(dpg.mvButton):
                dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
                dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 10, 6)
                dpg.add_theme_color(dpg.mvThemeCol_Button, (72, 78, 86))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, (88, 95, 104))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonActive, (98, 106, 116))
                dpg.add_theme_color(dpg.mvThemeCol_Text, COLOR_TEXT_DARK)

    if not dpg.does_item_exist("theme_input"):
        with dpg.theme(tag="theme_input"):
            for comp in (dpg.mvInputText, dpg.mvSliderInt, dpg.mvSliderFloat, dpg.mvCheckbox, dpg.mvCombo):
                with dpg.theme_component(comp):
                    dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
                    dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 8, 6)
                    dpg.add_theme_color(dpg.mvThemeCol_FrameBg, COLOR_BG_DARK_FRAME)
                    dpg.add_theme_color(dpg.mvThemeCol_FrameBgHovered, (58, 62, 64))
                    dpg.add_theme_color(dpg.mvThemeCol_FrameBgActive, (66, 70, 72))
                    dpg.add_theme_color(dpg.mvThemeCol_Text, COLOR_TEXT_DARK)

    if not dpg.does_item_exist("theme_dialog"):
        with dpg.theme(tag="theme_dialog"):
            with dpg.theme_component(dpg.mvFileDialog):
                dpg.add_theme_style(dpg.mvStyleVar_WindowPadding, SPACING, SPACING)
                dpg.add_theme_style(dpg.mvStyleVar_WindowRounding, RADIUS)
                dpg.add_theme_color(dpg.mvThemeCol_WindowBg, (20, 21, 24))
                dpg.add_theme_color(dpg.mvThemeCol_TitleBg, (18, 48, 42))
                dpg.add_theme_color(dpg.mvThemeCol_TitleBgActive, (24, 72, 63))

# ---------------- Layout helpers ----------------
def add_spacer(height: int = 10):
    dpg.add_spacer(height=height)

@contextmanager
def hgroup(spacing: int = SPACING):
    with dpg.group(horizontal=True, horizontal_spacing=spacing):
        yield

@contextmanager
def vgroup():
    with dpg.group(horizontal=False):
        yield

@contextmanager
def card(title: str, width: int = CARD_W, height: int = CARD_H, title_color=COLOR_TITLE):
    with dpg.child_window(width=width, height=height, border=True):
        dpg.add_text(title, color=title_color)
        add_spacer(6)
        yield

def tooltip(parent_item, text: str, color=None):
    if parent_item and dpg.does_item_exist(parent_item):
        with dpg.tooltip(parent_item):
            dpg.add_text(text) if not color else dpg.add_text(text, color=color)

def bullet(text: str, parent=None):
    msg = f"• {text}"
    return dpg.add_text(msg) if parent is None else dpg.add_text(msg, parent=parent)

def header_palette():
    return {
        "fg": (210, 230, 210),
        "accent": COLOR_NOTE,
        "muted": COLOR_MUTED,
    }

# ===================== Runtime accent themes =========================
# ملاحظة: هنستخدم tag فريد في كل مرة لتجنب "Alias already exists"
_RUNTIME_COUNTERS = {"primary": 0, "sidebar": 0}
_PRIMARY_THEME_CURRENT = "theme_button_primary"   # fallback to static baseline
_SIDEBAR_ACTIVE_THEME_CURRENT = None             # اختياري (مش مستخدم حالياً)

def get_primary_button_theme_tag() -> str:
    """Used by main_ui to bind active buttons (nav/current)."""
    return _PRIMARY_THEME_CURRENT or "theme_button_primary"

def _unique_tag(prefix: str) -> str:
    _RUNTIME_COUNTERS[prefix] += 1
    return f"__rt_{prefix}_btn_{_RUNTIME_COUNTERS[prefix]}"

def _mix(c1, c2, t):
    return int(c1 + (c2 - c1) * t)

def _lighten(rgb, amount=0.25):
    return tuple(_mix(c, 255, amount) for c in rgb)

def _darken(rgb, amount=0.35):
    return tuple(_mix(c, 0, amount) for c in rgb)

def _safe_delete(tag: str):
    try:
        if tag and dpg.does_item_exist(tag):
            dpg.delete_item(tag)
    except Exception:
        # لو حصل خطأ أثناء الحذف، نتجاهله ونطلع tag جديد لاحقاً
        pass

def set_accent_color(base_rgb, hover_rgb=None, active_rgb=None):
    """
    Rebuild PRIMARY runtime theme with a fresh unique tag.
    This avoids reusing an existing tag => no alias collision.
    """
    global _PRIMARY_THEME_CURRENT

    base = tuple(base_rgb)
    hov  = tuple(hover_rgb)  if hover_rgb  else _lighten(base, 0.25)
    act  = tuple(active_rgb) if active_rgb else _darken(base, 0.35)

    # حاول احذف الـ tag القديم (لو كان runtime)
    old_tag = _PRIMARY_THEME_CURRENT if _PRIMARY_THEME_CURRENT and _PRIMARY_THEME_CURRENT.startswith("__rt_") else None
    _safe_delete(old_tag)

    # ابنِ theme جديد بوسم فريد
    new_tag = _unique_tag("primary")
    with dpg.theme(tag=new_tag):
        with dpg.theme_component(dpg.mvButton):
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, RADIUS)
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 12, 8)
            dpg.add_theme_color(dpg.mvThemeCol_Button,        (*base, 255))
            dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, (*hov, 255))
            dpg.add_theme_color(dpg.mvThemeCol_ButtonActive,  (*act, 255))
            dpg.add_theme_color(dpg.mvThemeCol_Text,          (255, 255, 255))

    _PRIMARY_THEME_CURRENT = new_tag
    print(f"[THEME] Primary accent updated -> {new_tag} rgb={base}/{hov}/{act}")



import os
import shutil
import subprocess
import configparser
from datetime import datetime
import numpy as np
import pandas as pd

try:
    from plantvarfilter.linux import resolve_tool as _resolve_tool
except Exception:
    _resolve_tool = None


class HELPERS:
    COMMON_PLINK_FLAGS = ["--allow-extra-chr", "--chr-set", "26"]

    def replace_with_integers(self, input_file):
        mapping = {}
        current_integer = 1.0
        with open(input_file, 'r', errors="ignore") as infile:
            for line in infile:
                parts = line.strip().split('\t')
                col1_value = parts[0]
                try:
                    col1_value = int(col1_value)
                except ValueError:
                    if col1_value in mapping:
                        parts[0] = str(mapping[col1_value])
                    else:
                        mapping[col1_value] = current_integer
                        parts[0] = str(current_integer)
                        current_integer += 1
        return mapping

    def get_timestamp(self):
        now = datetime.now()
        dt_string = now.strftime("%d%m%Y_%H%M%S")
        return dt_string

    def save_raw_data(self, bed, pheno):
        np.save('snp', bed.read().val)
        np.savez_compressed('snp.npz', bed.read().val)
        np.save('pheno', pheno.read().val)

    def save_settings(self, default_path):
        config = configparser.ConfigParser()
        config['DefaultSettings'] = {'path': default_path, 'algorithm': 'FaST-LMM'}
        with open('settings.ini', 'w') as configfile:
            config.write(configfile)

    def get_settings(self, setting):
        config = configparser.ConfigParser()
        config.read('settings.ini')
        return config['DefaultSettings'][setting]

    def _safe_copy(self, src_path, dst_dir, add_log):
        if not src_path:
            return False
        if os.path.exists(src_path):
            try:
                os.makedirs(dst_dir, exist_ok=True)
                dst_path = os.path.join(dst_dir, os.path.basename(src_path))
                shutil.copy(src_path, dst_path)
                add_log(f"File saved: {os.path.basename(src_path)}")
                return True
            except Exception as e:
                add_log(f"[SAVE] Could not copy {src_path}: {e}", error=True)
        return False

    def _make_top_snps_csv(self, gwas_result_path, out_path, add_log, top_n=100):
        try:
            if not (gwas_result_path and os.path.exists(gwas_result_path)):
                return
            df = pd.read_csv(gwas_result_path)
            if 'PValue' in df.columns:
                df2 = df.dropna(subset=['PValue']).sort_values('PValue', ascending=True).head(top_n)
            elif 'SNP effect' in df.columns:
                df2 = df.dropna(subset=['SNP effect']).sort_values('SNP effect', ascending=False).head(top_n)
            else:
                df2 = df.head(top_n)
            df2.to_csv(out_path, index=False)
            add_log(f"File saved: {os.path.basename(out_path)}")
        except Exception as e:
            add_log(f"[TopSNPs] Failed to create top SNPs: {e}", warn=True)

    def save_results(
        self,
        current_dir,
        save_dir,
        gwas_result_name,
        gwas_result_name_top,
        manhatten_plot_name,
        qq_plot_name,
        algorithm,
        genomic_predict_name,
        gp_plot_name,
        gp_plot_name_scatter,
        add_log,
        settings_lst,
        pheno_stats_name,
        geno_stats_name
    ):
        def _is_writable(path: str) -> bool:
            try:
                if not path:
                    return False
                path = os.path.abspath(path)
                return os.path.isdir(path) and os.access(path, os.W_OK)
            except Exception:
                return False

        base_dir_candidates = [
            save_dir,
            current_dir,
            os.path.expanduser("~/Desktop"),
            os.path.expanduser("~")
        ]
        base_dir = None
        for cand in base_dir_candidates:
            if _is_writable(cand) and os.path.abspath(cand) != "/":
                base_dir = os.path.abspath(cand)
                break
        if base_dir is None:
            base_dir = os.getcwd()
            add_log(f"[SAVE] No writable directory found, falling back to {base_dir}", warn=True)
        else:
            if save_dir and os.path.abspath(save_dir) != base_dir:
                add_log(f"[SAVE] Selected folder not writable; using {base_dir} instead.", warn=True)

        safe_algo = str(algorithm or "Run").replace(" ", "_")
        ts = self.get_timestamp() + "_" + safe_algo
        dest_dir = os.path.join(base_dir, ts)
        try:
            os.makedirs(dest_dir, exist_ok=True)
        except Exception as e:
            add_log(f"Can not create folder '{dest_dir}': {e}", error=True)
            return base_dir

        def _abs_or_join(p):
            if not p:
                return None
            return p if os.path.isabs(p) else os.path.join(current_dir, os.path.basename(p))

        try:
            if gwas_result_name and os.path.exists(gwas_result_name):
                df = pd.read_csv(gwas_result_name)
                df.head(10000).to_csv(os.path.join(dest_dir, "gwas_results_top10000.csv"), index=False)
        except Exception as e:
            add_log(f"[SAVE] Could not create top 10000 GWAS CSV: {e}", warn=True)

        try:
            top_snps_path = os.path.join(dest_dir, "gwas_top_snps.csv")
            self._make_top_snps_csv(gwas_result_name, top_snps_path, add_log, top_n=100)
        except Exception as e:
            add_log(f"[TopSNPs] Could not generate: {e}", warn=True)

        manh_high = _abs_or_join(str(manhatten_plot_name or "").replace('manhatten_plot', 'manhatten_plot_high'))
        qq_high = _abs_or_join(str(qq_plot_name or "").replace('qq_plot', 'qq_plot_high'))
        gp_scatter_high = _abs_or_join(str(gp_plot_name_scatter or "").replace('GP_scatter_plot', 'GP_scatter_plot_high'))
        gp_ba_high = _abs_or_join(str(gp_plot_name or "").replace('Bland_Altman_plot', 'Bland_Altman_plot_high'))
        gp_validation = _abs_or_join(str(genomic_predict_name or "").replace('.csv', '_valdation.csv'))

        src_files = [
            _abs_or_join(gwas_result_name),
            _abs_or_join(genomic_predict_name),
            _abs_or_join(manhatten_plot_name),
            _abs_or_join(qq_plot_name),
            _abs_or_join(gp_plot_name),
            _abs_or_join(gp_plot_name_scatter),
            manh_high, qq_high, gp_scatter_high, gp_ba_high, gp_validation,
            _abs_or_join(pheno_stats_name),
            _abs_or_join(geno_stats_name),
        ]

        for src in src_files:
            self._safe_copy(src, dest_dir, add_log)

        try:
            log_path = os.path.join(dest_dir, 'log.txt')
            with open(log_path, 'w', encoding='utf-8') as log_file:
                def _get(i, default=""):
                    try:
                        return settings_lst[i]
                    except Exception:
                        return default

                log_file.write('Algorithm: ' + str(_get(0, 'N/A')))
                log_file.write('\nBed file used: ' + str(_get(1, 'N/A')))
                log_file.write('\nPheno file used: ' + str(_get(2, 'N/A')))
                try:
                    tr = 100 - (100 * float(_get(3, 0.3)))
                except Exception:
                    tr = 'N/A'
                log_file.write('\nTraining size: ' + str(tr))
                log_file.write('\nNr of trees: ' + str(_get(4, 'N/A')))
                log_file.write('\nNr of models: ' + str(_get(5, 'N/A')))
                log_file.write('\nMax depth: ' + str(_get(6, 'N/A')))
        except Exception as e:
            add_log(f"[SAVE] Could not write log.txt: {e}", warn=True)

        add_log(f"[INFO] Results saved in: {dest_dir}")
        return dest_dir

    def merge_gp_models(self, dataframes):
        df_combined = pd.concat(dataframes)
        df_result = df_combined.groupby(['ID1', 'BED_ID2'])['Predicted_Value'].mean().reset_index()
        df_result = pd.merge(dataframes[0], df_result, on='ID1', how='left')
        df_result = df_result.drop(['Predicted_Value_x', 'BED_ID2_y', 'Pheno_ID2'], axis=1)
        df_result = df_result.rename(columns={'Predicted_Value_y': 'Mean_Predicted_Value'})
        df_result['Difference'] = (df_result['Pheno_Value'] - df_result['Mean_Predicted_Value']).abs()
        df_result['Difference'] = df_result['Difference'].round(decimals=3)
        df_result['Mean_Predicted_Value'] = df_result['Mean_Predicted_Value'].round(decimals=3)
        return df_result

    def merge_models(self, dataframes, method):
        df_combined = pd.concat(dataframes)
        df_combined = df_combined[df_combined['PValue'] > 0]
        df_grouped = df_combined.groupby(['SNP', 'Chr', 'ChrPos'])['PValue'].agg([method, 'std']).reset_index()
        df_grouped = df_grouped.rename(columns={method: 'PValue_x', 'std': 'PValue_sd'})
        df_result = df_grouped.rename(columns={'PValue_x': 'PValue'})
        df_result['Chr'] = df_result['Chr'].astype(int)
        df_result['ChrPos'] = df_result['ChrPos'].astype(int)
        df_result = df_result.sort_values(by=['Chr', 'ChrPos'])
        return df_result

    @staticmethod
    def plink_bin() -> str:
        if '_resolve_tool' in globals() and _resolve_tool is not None:
            p = _resolve_tool("plink")
            if p:
                return p
        p = shutil.which("plink")
        return p if p else "plink"

    def run_plink(self, cmd, log=print):
        log(" ".join(cmd))
        p = subprocess.run(cmd, capture_output=True, text=True)
        if p.stdout:
            for ln in p.stdout.strip().splitlines():
                log(ln)
        if p.returncode != 0:
            if p.stderr:
                log(p.stderr.strip())
            raise RuntimeError("plink failed with exit code {}".format(p.returncode))



# variant_caller_utils.py
import os
import shutil
import subprocess
import tempfile
from typing import List, Optional, Tuple, Callable, Union, Dict

from bcftools_utils import BCFtools

try:
    from plantvarfilter.linux import resolve_tool
except Exception:
    resolve_tool = None

LogFn = Callable[[str], None]


class VariantCallerError(RuntimeError):
    pass


def _resolve_exe(name: str) -> str:
    if resolve_tool is not None:
        p = resolve_tool(name)
        if p:
            return p
    p = shutil.which(name)
    if p:
        return p
    raise VariantCallerError(f"Executable not found: {name}. Bundle it under PlantVarFilter/linux or add to PATH.")


class VariantCaller:
    def __init__(
        self,
        bcftools_bin: Optional[str] = None,
        bgzip_bin: Optional[str] = None,
        tabix_bin: Optional[str] = None,
    ):
        self.bcftools = bcftools_bin or _resolve_exe("bcftools")
        self.bgzip = bgzip_bin or _resolve_exe("bgzip")
        self.tabix = tabix_bin or _resolve_exe("tabix")
        self.last_split: Optional[Dict[str, str]] = None

    def _run(self, cmd: List[str], log: LogFn):
        log(" ".join(cmd))
        res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        out = res.stdout.decode("utf-8", errors="replace")
        if res.returncode != 0:
            log(out)
            raise VariantCallerError(f"Command failed: {' '.join(cmd)}")
        if out.strip():
            for ln in out.strip().splitlines():
                log(ln)

    def get_last_split(self) -> Optional[Dict[str, str]]:
        return self.last_split

    def call_bcftools(
        self,
        bams: Union[List[str], str],
        ref_fasta: str,
        out_prefix: Optional[str] = None,
        regions_bed: Optional[str] = None,
        threads: int = 8,
        min_baseq: int = 13,
        min_mapq: int = 20,
        ploidy: int = 2,
        log: LogFn = print,
        split_after_calling: bool = False,
    ) -> Tuple[str, str]:
        if not os.path.exists(ref_fasta):
            raise VariantCallerError(f"Reference FASTA not found: {ref_fasta}")

        workdir = os.getcwd()
        if isinstance(bams, list) and len(bams) > 0:
            workdir = os.path.dirname(os.path.abspath(bams[0])) or workdir
        elif isinstance(bams, str) and os.path.exists(bams):
            workdir = os.path.dirname(os.path.abspath(bams)) or workdir

        if not out_prefix:
            out_prefix = os.path.join(workdir, "calls")

        vcf_gz = f"{out_prefix}.vcf.gz"
        out_dir = os.path.dirname(vcf_gz) or "."
        os.makedirs(out_dir, exist_ok=True)

        bam_list_path = None
        single_bam: Optional[str] = None
        is_temp_bamlist = False

        if isinstance(bams, list):
            if len(bams) == 0:
                raise VariantCallerError("No BAMs provided.")
            if len(bams) == 1:
                single_bam = bams[0]
            else:
                fd, tmp = tempfile.mkstemp(prefix="bamlist_", suffix=".list", dir=workdir)
                os.close(fd)
                with open(tmp, "w") as fh:
                    for b in bams:
                        fh.write(b + "\n")
                bam_list_path = tmp
                single_bam = None
                is_temp_bamlist = True
        else:
            if not os.path.exists(bams):
                raise VariantCallerError(f"BAM list not found: {bams}")
            bam_list_path = bams
            single_bam = None
            is_temp_bamlist = False

        try:
            mp_cmd = [
                self.bcftools,
                "mpileup",
                "-Ou",
                "-f",
                ref_fasta,
                "-q",
                str(min_mapq),
                "-Q",
                str(min_baseq),
                "-a",
                "FORMAT/AD,FORMAT/DP",
            ]
            if threads and threads > 1:
                mp_cmd += ["--threads", str(threads)]
            if regions_bed and os.path.exists(regions_bed):
                mp_cmd += ["-R", regions_bed]
            if bam_list_path:
                mp_cmd += ["-b", bam_list_path]
            elif single_bam:
                mp_cmd += [single_bam]

            call_cmd = [
                self.bcftools,
                "call",
                "-mv",
                "--ploidy",
                str(ploidy),
                "-Oz",
                "-o",
                vcf_gz,
            ]
            if threads and threads > 1:
                call_cmd = ["taskset", "-c", f"0-{max(0, threads - 1)}"] + call_cmd

            log(" | ".join([" ".join(mp_cmd), " ".join(call_cmd)]))
            p1 = subprocess.Popen(mp_cmd, stdout=subprocess.PIPE)
            p2 = subprocess.Popen(call_cmd, stdin=p1.stdout, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            p1.stdout.close()
            out = p2.communicate()[0]
            rc = p2.returncode
            if rc != 0:
                out_txt = (out or b"").decode("utf-8", errors="replace")
                raise VariantCallerError(f"bcftools call failed:\n{out_txt}")

            self._run([self.tabix, "-f", "-p", "vcf", vcf_gz], log)

            self.last_split = None
            if split_after_calling:
                try:
                    bcf = BCFtools()
                    paths = bcf.split_vcf(
                        input_vcf=vcf_gz,
                        out_prefix=out_prefix,
                        log=log,
                        regions_bed=regions_bed,
                    )
                    self.last_split = paths
                    if paths.get("snps"):
                        log(f"SNPs VCF: {paths['snps']}")
                    if paths.get("indels"):
                        log(f"INDELs VCF: {paths['indels']}")
                except Exception as e:
                    log(f"WARN: split_vcf failed: {e}")

            return vcf_gz, vcf_gz + ".tbi"
        finally:
            if is_temp_bamlist and bam_list_path and os.path.exists(bam_list_path):
                try:
                    os.remove(bam_list_path)
                except Exception:
                    pass



# bigfile_handler.py
import os
import io
import gzip
import json
import math
import shutil
import hashlib
import tempfile
import subprocess
from dataclasses import dataclass
from typing import Callable, Iterator, List, Optional, Tuple

# -----------------------------
# Settings and small utilities
# -----------------------------

@dataclass
class LargeFileSettings:
    enabled: bool = True
    # Split by number of variant lines per chunk (header lines are copied to each chunk)
    chunk_lines: int = 500_000
    # Max number of parallel workers when mapping chunks
    max_workers: int = 2
    # Optional temp directory (falls back to system tmp if None)
    temp_dir: Optional[str] = None
    # Resume from a previous run (uses checkpoint file under temp_dir)
    resume: bool = True
    # Merge method for VCFs: "bcftools" or "cat" (bcftools is safer for VCFs)
    merge_strategy: str = "bcftools"
    # Gzip-level for chunk files created by Python writer (0-9)
    compress_level: int = 5
    # Buffer size in bytes for streaming
    io_buffer_bytes: int = 2 * 1024 * 1024

class BigFileError(RuntimeError):
    pass

class TempManager:
    def __init__(self, base: Optional[str] = None, prefix: str = "pvf_"):
        self._root = base or tempfile.gettempdir()
        os.makedirs(self._root, exist_ok=True)
        self._dir = tempfile.mkdtemp(prefix=prefix, dir=self._root)
        self._alive = True

    @property
    def path(self) -> str:
        return self._dir

    def cleanup(self):
        if self._alive:
            try:
                shutil.rmtree(self._dir, ignore_errors=True)
            finally:
                self._alive = False

    def __del__(self):
        try:
            self.cleanup()
        except Exception:
            pass

class Checkpoint:
    def __init__(self, path: str):
        self.path = path
        self.state = {"done": [], "meta": {}}
        self._load()

    def _load(self):
        try:
            if os.path.exists(self.path):
                with open(self.path, "r", encoding="utf-8") as f:
                    self.state = json.load(f)
        except Exception:
            self.state = {"done": [], "meta": {}}

    def save(self):
        try:
            with open(self.path, "w", encoding="utf-8") as f:
                json.dump(self.state, f, ensure_ascii=False, indent=2)
        except Exception:
            pass

    def mark_done(self, key: str):
        if key not in self.state["done"]:
            self.state["done"].append(key)
            self.save()

    def is_done(self, key: str) -> bool:
        return key in self.state["done"]

class ExternalRunner:
    @staticmethod
    def run(cmd: List[str], log: Callable[[str], None], check: bool = True) -> int:
        log(" ".join(cmd))
        p = subprocess.run(cmd, text=True, capture_output=True)
        if p.stdout:
            for ln in p.stdout.splitlines():
                log(ln)
        if p.stderr:
            for ln in p.stderr.splitlines():
                log(ln)
        if check and p.returncode != 0:
            raise BigFileError(f"Command failed: {' '.join(cmd)} (rc={p.returncode})")
        return p.returncode

# --------------------------------
# Core: BigFileProcessor (generic)
# --------------------------------

class BigFileProcessor:
    """
    Generic large-file helper for:
      - Streaming and chunking text files (VCF/VCF.GZ)
      - Mapping a function over chunks (in sequence or with limited parallelism)
      - Merging chunk outputs (prefer bcftools for VCF)
      - Checkpoint/Resume
    """
    def __init__(self, settings: LargeFileSettings, logger: Callable[[str], None] = print):
        self.s = settings
        self.log = logger
        self.tmp = TempManager(self.s.temp_dir, prefix="pvf_big_")

    # ---------- Basics ----------

    @staticmethod
    def _is_gzip(path: str) -> bool:
        return path.lower().endswith(".gz")

    @staticmethod
    def _hash_path(path: str) -> str:
        h = hashlib.sha1(path.encode("utf-8", errors="ignore")).hexdigest()[:10]
        return h

    def _open_text(self, path: str):
        # Returns a file-like object in text mode for .gz or plain
        if self._is_gzip(path):
            return io.TextIOWrapper(
                gzip.open(path, "rb"),
                encoding="utf-8",
                newline=""
            )
        return open(path, "r", encoding="utf-8", newline="")

    # ---------- VCF helpers ----------

    def _read_vcf_header(self, path: str) -> List[str]:
        header = []
        with self._open_text(path) as f:
            for line in f:
                if line.startswith("#"):
                    header.append(line)
                else:
                    break
        if not header or not header[-1].startswith("#CHROM"):
            # Fallback: try to re-scan to guarantee #CHROM exists
            with self._open_text(path) as f:
                for line in f:
                    if line.startswith("#CHROM"):
                        header.append(line)
                        break
        if not header:
            raise BigFileError("Invalid VCF: missing header lines.")
        return header

    def _iter_vcf_records(self, path: str) -> Iterator[str]:
        with self._open_text(path) as f:
            for line in f:
                if line and (line[0] != "#"):
                    yield line

    # ---------- Chunking ----------

    def chunk_vcf_by_lines(self, input_vcf: str, chunk_lines: Optional[int] = None) -> List[str]:
        """
        Split a VCF/VCF.GZ into N chunk files (gzipped), copying header to each chunk.
        Returns a list of chunk paths.
        """
        clines = int(chunk_lines or self.s.chunk_lines)
        header = self._read_vcf_header(input_vcf)
        base = os.path.basename(input_vcf)
        stem = base.replace(".vcf.gz", "").replace(".vcf", "")
        out_paths: List[str] = []
        idx = 0
        written = 0

        def _new_chunk_path(i: int) -> str:
            return os.path.join(self.tmp.path, f"{stem}.part{i:05d}.vcf.gz")

        writer = None
        gz = None

        try:
            for rec in self._iter_vcf_records(input_vcf):
                if writer is None:
                    part_path = _new_chunk_path(idx)
                    gz = gzip.open(part_path, "wb", compresslevel=self.s.compress_level)
                    writer = io.TextIOWrapper(gz, encoding="utf-8")
                    for h in header:
                        writer.write(h)
                    out_paths.append(part_path)
                    written = 0
                writer.write(rec)
                written += 1
                if written >= clines:
                    writer.flush()
                    writer.detach()
                    gz.close()
                    writer = None
                    gz = None
                    idx += 1
            # Close last
            if writer is not None:
                writer.flush()
                writer.detach()
                gz.close()
        finally:
            try:
                if writer is not None:
                    writer.detach()
                if gz is not None:
                    gz.close()
            except Exception:
                pass

        self.log(f"[BigFile] Chunked VCF: {len(out_paths)} part(s)")
        return out_paths

    # ---------- Merge ----------

    def merge_vcfs(self, vcfs: List[str], out_vcf_gz: str, bcftools_path: Optional[str] = "bcftools") -> str:
        if not vcfs:
            raise BigFileError("No VCFs to merge.")
        os.makedirs(os.path.dirname(out_vcf_gz) or ".", exist_ok=True)

        if self.s.merge_strategy.lower() == "bcftools":
            # Safe merge using bcftools concat -a
            cmd = [bcftools_path or "bcftools", "concat", "-a", "-Oz", "-o", out_vcf_gz] + vcfs
            ExternalRunner.run(cmd, self.log, check=True)
            ExternalRunner.run([bcftools_path or "bcftools", "index", "-f", "-p", "vcf", out_vcf_gz], self.log, check=False)
            return out_vcf_gz

        # Fallback simple concatenation (assumes compatible headers)
        # Extract header from first file + all body lines from each
        header = self._read_vcf_header(vcfs[0])
        with gzip.open(out_vcf_gz, "wb", compresslevel=self.s.compress_level) as gz:
            w = io.TextIOWrapper(gz, encoding="utf-8")
            for h in header:
                w.write(h)
            for path in vcfs:
                with self._open_text(path) as f:
                    for line in f:
                        if line and (line[0] != "#"):
                            w.write(line)
            w.flush()
        return out_vcf_gz

    # ---------- Map over chunks ----------

    def map_chunks(
        self,
        chunk_paths: List[str],
        fn: Callable[[str], str],
        checkpoint_key: str,
        checkpoint_dir: Optional[str] = None
    ) -> List[str]:
        """
        Apply fn(chunk_path) -> output_path on each chunk sequentially (worker pool can be added later).
        Respects resume when enabled.
        """
        ck_dir = checkpoint_dir or self.tmp.path
        os.makedirs(ck_dir, exist_ok=True)
        ck_path = os.path.join(ck_dir, f"{checkpoint_key}.json")
        ck = Checkpoint(ck_path)

        outputs: List[str] = []
        total = len(chunk_paths)
        for i, part in enumerate(chunk_paths, 1):
            key = os.path.basename(part)
            if self.s.resume and ck.is_done(key):
                self.log(f"[BigFile] Skip (resume): {key} ({i}/{total})")
                # Assume output present next to chunk under naming rule:
                out_guess = self._guess_output_path(part)
                if out_guess and os.path.exists(out_guess):
                    outputs.append(out_guess)
                continue
            self.log(f"[BigFile] Processing: {key} ({i}/{total})")
            out = fn(part)
            outputs.append(out)
            ck.mark_done(key)
        return outputs

    @staticmethod
    def _guess_output_path(chunk_path: str) -> Optional[str]:
        # Heuristic: replace ".partXXXXX.vcf.gz" with ".partXXXXX.proc.vcf.gz"
        if chunk_path.endswith(".vcf.gz"):
            return chunk_path[:-7] + ".proc.vcf.gz"
        return None

    # ---------- High-level pipelines ----------

    def preprocess_vcf_in_chunks(
        self,
        input_vcf: str,
        out_prefix: str,
        build_chunk_cmd: Callable[[str, str], List[str]],
        bcftools_path: Optional[str] = "bcftools",
    ) -> Tuple[str, List[str]]:
        """
        Generic chunked preprocessing:
          - split input into chunks
          - run external command per chunk (provided by build_chunk_cmd)
          - merge outputs with bcftools concat
        build_chunk_cmd(part_vcf, out_vcf_gz) -> command list
        """
        chunks = self.chunk_vcf_by_lines(input_vcf, self.s.chunk_lines)

        def _apply(part: str) -> str:
            out_part = part[:-7] + ".proc.vcf.gz" if part.endswith(".vcf.gz") else part + ".proc.vcf.gz"
            cmd = build_chunk_cmd(part, out_part)
            ExternalRunner.run(cmd, self.log, check=True)
            # index per-part to speed up concat
            ExternalRunner.run([bcftools_path or "bcftools", "index", "-f", "-p", "vcf", out_part], self.log, check=False)
            return out_part

        outputs = self.map_chunks(chunks, _apply, checkpoint_key=f"ck_{self._hash_path(input_vcf)}")
        final_out = out_prefix + ".vcf.gz" if not out_prefix.endswith(".vcf.gz") else out_prefix
        merged = self.merge_vcfs(outputs, final_out, bcftools_path=bcftools_path)
        self.log(f"[BigFile] Final merged: {merged}")
        return merged, outputs



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
import xgboost as xgb
from sklearn.ensemble import RandomForestRegressor

###This file is depreciated
class GWASAI:
    """This file is depreciated."""
    def run_random_forest(self, snp_data, pheno_data, df_bim, test_size, estimators, gwas_result_name, bed_gp, pheno_gp,
                    genomic_predict, genomic_predict_name, model_nr):
        print(np.sum(np.isnan(snp_data)))
        snp_data[np.isnan(snp_data)] = -1
        print(np.sum(np.isnan(snp_data)))
        print(snp_data.shape, pheno_data.shape)
        # Split data into training and testing sets
        X_train, X_test, y_train, y_test = train_test_split(snp_data, pheno_data, test_size=test_size, random_state=42)#, random_state=42)
        print(X_train.shape, X_test.shape, y_train.shape, y_test.shape)
        # # Standardize the input features
        ### for GP scalling might be not good
        if genomic_predict:
            pass
        else:
            #print ('scaling')
            scaler = StandardScaler()
            X_train = scaler.fit_transform(X_train)
        #X_test = scaler.transform(X_test)

        # Define and train a Random Forest model
        rf_model = RandomForestRegressor(n_estimators=estimators, n_jobs=-1, random_state=42)#, random_state=0) #42
        #rf_model = RandomForestClassifier(n_estimators=estimators,max_depth=2, random_state=0)
        #print(X_train, y_train, y_test)
        #print (type(X_train), type(y_train))
        #import matplotlib.pyplot as plt
        #import seaborn as sns
        #plt.imshow(X_train, cmap='viridis')  # 'cmap' denotes the color map
        #plt.colorbar()  # Optional: adds a color bar to the side
        #plt.show()
        #df = pd.DataFrame(X_train)
        # Create a heatmap
        #sns.heatmap(df)
        #plt.show()

        # Fit the model to the training data
        rf_model.fit(X_train, y_train)
        print(rf_model.feature_importances_)
        # import seaborn as sns
        # print(len(snp_data))
        #
        # importances = rf_model.feature_importances_
        # indices = np.argsort(importances)[::-1]
        #
        # # Get the indices of the top 10 features
        # top_10_indices = np.argsort(importances)[-50:]
        #
        # # Create a DataFrame with the top 10 SNPs
        # top_10_snps = pd.DataFrame(snp_data[:, top_10_indices], columns=[f'SNP_{i}' for i in top_10_indices])
        #
        # # Calculate the correlation matrix
        # corr = top_10_snps.corr()
        #
        # # Generate a heatmap
        # plt.figure(figsize=(10, 8))
        # sns.heatmap(corr, annot=True, cmap='coolwarm', fmt=".2f", xticklabels=1, yticklabels=1)
        # plt.title("Correlation Heatmap of Top 10 SNPs")
        # plt.show()

        if genomic_predict:
            bed_data = bed_gp.read().iid
            pheno_data = pheno_gp.read().iid
            df_gp = pd.DataFrame(bed_data, columns=['ID1', 'BED_ID2'])
            predicted_values = rf_model.predict(bed_gp.read().val)
            df_gp['Predicted_Value'] = predicted_values
            #print (df_gp)
            df_pheno = pd.DataFrame(pheno_data, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno_gp.read().val

            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Difference'] = (merged_df['Pheno_Value'] - merged_df['Predicted_Value']).abs()
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(3)
            merged_df['Difference'] = merged_df['Difference'].round(3)
            merged_df.to_csv(genomic_predict_name, sep=';', index=False)
            return merged_df

        else:

            # features_dict = dict(zip(snp_ids, rf_model.feature_importances_))
            # # Sort the dictionary by importance
            # sorted_features = sorted(features_dict.items(), key=lambda x: x[1], reverse=True)
            # # Convert to a DataFrame for easy saving
            # sorted_features_df = pd.DataFrame(sorted_features, columns=['Feature', 'Importance'])
            # # Save to a CSV file
            # sorted_features_df.to_csv(gwas_result_name, index=False)
            # sorted_features_df.columns = ['SNP', 'PValue']
            # sorted_features_df[['Chr', 'ChrPos']] = sorted_features_df['SNP'].str.split(':', expand=True)

            # f = open(gwas_result_name, 'w')

            data = []
            snp_ids = df_bim.iloc[:, 1].tolist()
            for col, score in zip(snp_ids, rf_model.feature_importances_):
                data.append((col, score))
            # Convert the list of tuples into a DataFrame
            df = pd.DataFrame(data, columns=['SNP', 'PValue'])
            df = pd.merge(df, df_bim, on='SNP', how='left')
            del df['NA']
            #df[['Chr', 'ChrPos']] = df['SNP'].str.split(':', expand=True)
            return df

    def run_xgboost(self, snp_data, pheno_data, df_bim, test_size, estimators, gwas_result_name, bed_gp, pheno_gp,
                    genomic_predict, genomic_predict_name, max_dep_set, model_nr):

        snp_data[np.isnan(snp_data)] = -1

        # Split data into training and testing sets
        X_train, X_test, y_train, y_test = train_test_split(snp_data, pheno_data, test_size=test_size)#, random_state=42)

        if genomic_predict:
            pass
        else:

            scaler = StandardScaler()
            X_train = scaler.fit_transform(X_train)
        # # Standardize the input features
        #scaler = StandardScaler()
        #X_train = scaler.fit_transform(X_train)
        #X_test = scaler.transform(X_test)

        # Define and train a Random Forest model
        xgb_model = xgb.XGBRegressor(n_estimators=estimators, learning_rate=0.1, max_depth=max_dep_set)#, random_state=0)
        #xgb_model = xgb.XGBRegressor(n_estimators=estimators, random_state=42)

        # Fit the model to the training data
        xgb_model.fit(X_train, y_train)

        # test_predictions = rf_model.predict(X_test)
        # test_loss = np.mean((test_predictions - y_test)**2)  # Mean squared error
        # print(f'Test Loss: {test_loss}')

        if genomic_predict:
            bed_data = bed_gp.read().iid
            pheno_data = pheno_gp.read().iid
            df_gp = pd.DataFrame(bed_data, columns=['ID1', 'BED_ID2'])
            predicted_values = xgb_model.predict(bed_gp.read().val)
            df_gp['Predicted_Value'] = predicted_values

            df_pheno = pd.DataFrame(pheno_data, columns=['ID1', 'Pheno_ID2'])
            df_pheno['Pheno_Value'] = pheno_gp.read().val
            merged_df = df_gp.merge(df_pheno, on='ID1', how='outer')
            merged_df['Difference'] = (merged_df['Pheno_Value'] - merged_df['Predicted_Value']).abs()
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].astype(float)
            merged_df['Predicted_Value'] = merged_df['Predicted_Value'].round(3)
            merged_df['Difference'] = merged_df['Difference'].round(3)

            merged_df.to_csv(genomic_predict_name, sep=';', index=False)
            return merged_df

        else:
            # features_dict = dict(zip(snp_ids, xgb_model.feature_importances_))
            # # Sort the dictionary by importance
            # sorted_features = sorted(features_dict.items(), key=lambda x: x[1], reverse=True)
            # # Convert to a DataFrame for easy saving
            # sorted_features_df = pd.DataFrame(sorted_features, columns=['Feature', 'Importance'])
            # # Save to a CSV file
            # sorted_features_df.to_csv(gwas_result_name, index=False)
            # sorted_features_df.columns = ['SNP', 'PValue']
            # sorted_features_df[['Chr', 'ChrPos']] = sorted_features_df['SNP'].str.split(':', expand=True)

            #f = open(gwas_result_name, 'w')
            data = []
            snp_ids = df_bim.iloc[:, 1].tolist()
            for col, score in zip(snp_ids, xgb_model.feature_importances_):
                #f.write(str(col) + ' ' + str(score) + '\n')
                data.append((col, score))
            # Convert the list of tuples into a DataFrame
            df = pd.DataFrame(data, columns=['SNP', 'PValue'])
            df = pd.merge(df, df_bim, on='SNP', how='left')
            del df['NA']
            #df[['Chr', 'ChrPos']] = df['SNP'].str.split(':', expand=True)
            return df




# plantvarfilter/batch_gwas.py
import os
import tempfile
import pandas as pd
from pysnptools.snpreader import Bed, Pheno
import pysnptools.util as pstutil

def _read_pheno_multi(pheno_path):
    try:
        df = pd.read_csv(pheno_path)
        if df.shape[1] < 3:
            raise ValueError
        return df
    except Exception:
        df = pd.read_csv(pheno_path, sep=r"\s+|,", engine="python", header=0)
        if df.shape[1] < 3:
            raise ValueError(f"Phenotype file must have at least FID IID and one trait column: {pheno_path}")
        return df

def run_batch_gwas_for_all_traits(
    gwas, helper, bed_path, pheno_path, cov_path, algorithm, out_dir,
    log_fn, nr_jobs=-1, gb_goal=0, train_size=0.3, estimators=200, model_nr=1,
    max_depth=3, aggregation_method="sum", snp_limit=None
):
    os.makedirs(out_dir, exist_ok=True)
    dfp = _read_pheno_multi(pheno_path)

    cols = list(dfp.columns)
    fid_col, iid_col = cols[0], cols[1]
    trait_cols = cols[2:]
    if len(trait_cols) == 0:
        raise ValueError("No trait columns found. Expecting: FID, IID, <Trait1>, <Trait2>, ...")

    chrom_mapping = helper.replace_with_integers(bed_path.replace(".bed", ".bim"))

    summary_rows = []
    per_trait_outputs = []

    bed = Bed(str(bed_path), count_A1=False, chrom_map=chrom_mapping)

    cov = Pheno(str(cov_path)) if cov_path else None

    with tempfile.TemporaryDirectory() as tmpdir:
        for trait in trait_cols:
            trait_pheno_path = os.path.join(tmpdir, f"pheno_{trait}.txt")
            sub = dfp[[fid_col, iid_col, trait]].copy()
            sub.columns = ["FID", "IID", "Value"]
            sub = sub.dropna(subset=["Value"])
            sub.to_csv(trait_pheno_path, sep="\t", index=False, header=False)

            pheno = Pheno(str(trait_pheno_path))

            bed_i, pheno_i = pstutil.intersect_apply([bed, pheno])
            bed_fixed = gwas.filter_out_missing(bed_i)

            log_fn(f"[Batch-GWAS] Trait '{trait}': N_SNP={bed_fixed.sid_count}, N_samples={pheno_i.iid_count}")

            res_csv = os.path.join(out_dir, f"{trait}_gwas_results.csv")
            man_png = os.path.join(out_dir, f"{trait}_manhattan.png")
            qq_png  = os.path.join(out_dir, f"{trait}_qq.png")

            gwas_df, df_plot = None, None
            if algorithm in ("FaST-LMM", "Linear regression"):
                gwas_df, df_plot = gwas.run_gwas_lmm(
                    bed_fixed, pheno_i, chrom_mapping, log_fn,
                    res_csv, algorithm, bed_path, cov, gb_goal
                )
            elif algorithm == "Random Forest (AI)":
                gwas_df, df_plot = gwas.run_gwas_rf(
                    bed_fixed, pheno_i, bed_path, 1.0 - train_size, estimators,
                    res_csv, chrom_mapping, log_fn, model_nr, nr_jobs, aggregation_method
                )
            elif algorithm == "XGBoost (AI)":
                gwas_df, df_plot = gwas.run_gwas_xg(
                    bed_fixed, pheno_i, bed_path, 1.0 - train_size, estimators,
                    res_csv, chrom_mapping, log_fn, model_nr, max_depth, nr_jobs, aggregation_method
                )
            elif algorithm == "Ridge Regression":
                gwas_df, df_plot = gwas.run_gwas_ridge(
                    bed_fixed, pheno_i, bed_path, 1.0 - train_size, 1.0,
                    res_csv, chrom_mapping, log_fn, model_nr, aggregation_method
                )
            else:
                raise ValueError(f"Unsupported algorithm for batch mode: {algorithm}")

            if gwas_df is None:
                log_fn(f"[Batch-GWAS] Trait '{trait}': no results produced", warn=True)
                continue

            try:
                gwas.plot_gwas(df_plot, snp_limit, algorithm, man_png, qq_png, chrom_mapping)
            except Exception:
                pass

            top_snp, top_score = None, None
            if algorithm in ("FaST-LMM", "Linear regression") and "PValue" in gwas_df.columns:
                gwas_df = gwas_df.sort_values("PValue", ascending=True)
                row0 = gwas_df.iloc[0]
                top_snp, top_score = str(row0.get("SNP")), float(row0.get("PValue"))
            else:
                colname = "SNP effect"
                if colname in gwas_df.columns:
                    gwas_df = gwas_df.sort_values(colname, ascending=False)
                    row0 = gwas_df.iloc[0]
                    top_snp, top_score = str(row0.get("SNP")), float(row0.get(colname))

            summary_rows.append({
                "Trait": trait,
                "Top_hit": top_snp,
                "Score": top_score,
                "ResultsCSV": os.path.basename(res_csv),
                "ManhattanPNG": os.path.basename(man_png) if os.path.exists(man_png) else "",
                "QQPNG": os.path.basename(qq_png) if os.path.exists(qq_png) else "",
            })
            per_trait_outputs.append({
                "trait": trait,
                "csv": res_csv,
                "manhattan": man_png if os.path.exists(man_png) else None,
                "qq": qq_png if os.path.exists(qq_png) else None,
            })
            log_fn(f"[Batch-GWAS] Trait '{trait}': done → {os.path.basename(res_csv)}")

    summary_path = os.path.join(out_dir, "batch_gwas_summary.csv")
    pd.DataFrame(summary_rows).to_csv(summary_path, index=False)
    return {
        "summary_csv": summary_path,
        "per_trait": per_trait_outputs
    }



# plantvarfilter/preanalysis/reference_manager.py

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import os
import shutil
import subprocess
from typing import Dict, Optional


# Prefer the project's resolver if available
try:
    from plantvarfilter.variant_caller_utils import resolve_tool  # type: ignore
except Exception:
    resolve_tool = shutil.which  # fallback


@dataclass
class ReferenceIndexStatus:
    """Summary of reference indexing artifacts and tool availability."""
    fasta: str
    reference_dir: str
    faidx: Optional[str]
    dict: Optional[str]
    mmi: Optional[str]
    bt2_prefix: Optional[str]
    tools: Dict[str, Dict[str, Optional[str]]]
    ok: bool


class ReferenceManager:
    """
    Build and validate reference indexes used by downstream steps.

    Artifacts (co-located in a dedicated 'reference' directory):
      - FASTA copy/link
      - <fasta>.fai                  (samtools faidx)
      - <fasta_stem>.dict            (samtools dict or Picard)
      - <fasta_stem>.mmi             (minimap2 index)
      - <fasta_stem>_bt2.*.bt2[l]?   (bowtie2-build index)
    """

    def __init__(self, logger=print, workspace: Optional[str] = None):
        self.log = logger
        self.workspace = Path(workspace) if workspace else None

    # ---------------------------- public API ---------------------------- #

    def build_indices(
        self,
        fasta: str,
        out_dir: Optional[str] = None,
        build_mmi: bool = True,
        build_bt2: bool = True,
        build_dict: bool = True,
    ) -> ReferenceIndexStatus:
        """
        Build any missing indexes for the given FASTA.
        Returns a ReferenceIndexStatus describing what exists.
        """
        fa_src = Path(fasta).expanduser().resolve()
        if not fa_src.exists():
            raise FileNotFoundError(f"FASTA not found: {fa_src}")

        ref_dir = Path(out_dir or self._default_reference_dir_for(fa_src)).resolve()
        ref_dir.mkdir(parents=True, exist_ok=True)

        # Place a copy/hardlink of the FASTA inside reference dir so sidecar files sit together.
        fa_dst = ref_dir / fa_src.name
        if not fa_dst.exists():
            try:
                os.link(str(fa_src), str(fa_dst))  # hard link when possible
                self.log(f"[REF] Linked FASTA → {fa_dst}")
            except Exception:
                shutil.copy2(str(fa_src), str(fa_dst))
                self.log(f"[REF] Copied FASTA → {fa_dst}")

        # Build faidx
        samtools = self._tool_path("samtools")
        if samtools:
            fai = fa_dst.with_suffix(fa_dst.suffix + ".fai")
            if not fai.exists():
                self._run([samtools, "faidx", str(fa_dst)], tag="samtools")
        else:
            self.log("[REF] samtools not found in PATH (faidx will be missing)")

        # Build dict (samtools dict preferred; Picard fallback if available)
        if build_dict:
            dict_path = ref_dir / (fa_dst.stem + ".dict")
            if not dict_path.exists():
                if samtools and self._supports_samtools_dict(samtools):
                    try:
                        self._run([samtools, "dict", str(fa_dst), "-o", str(dict_path)], tag="samtools")
                    except Exception as e:
                        self.log(f"[REF] samtools dict failed ({e}). Will try Picard if present.")
                        self._try_picard_dict(fa_dst, dict_path)
                else:
                    self._try_picard_dict(fa_dst, dict_path)

        # Build minimap2 .mmi
        if build_mmi:
            minimap2 = self._tool_path("minimap2")
            mmi = ref_dir / (fa_dst.stem + ".mmi")
            if minimap2:
                if not mmi.exists():
                    self._run([minimap2, "-d", str(mmi), str(fa_dst)], tag="minimap2")
            else:
                self.log("[REF] minimap2 not found in PATH (mmi will be missing)")

        # Build bowtie2 index
        if build_bt2:
            bt2_build = self._tool_path("bowtie2-build")
            bt2_prefix = ref_dir / (fa_dst.stem + "_bt2")
            if bt2_build:
                if not self._bowtie2_index_exists(bt2_prefix):
                    self._run([bt2_build, str(fa_dst), str(bt2_prefix)], tag="bowtie2-build")
            else:
                self.log("[REF] bowtie2-build not found in PATH (bt2 index will be missing)")

        return self.check_status(str(fa_dst))

    def check_status(self, fasta_in_reference_dir: str) -> ReferenceIndexStatus:
        """Return a status snapshot of all reference artifacts and tools."""
        fa = Path(fasta_in_reference_dir).expanduser().resolve()
        ref_dir = fa.parent

        tools = {
            "samtools": self._tool_info("samtools"),
            "minimap2": self._tool_info("minimap2"),
            "bowtie2-build": self._tool_info("bowtie2-build"),
            "picard": self._tool_info("picard"),  # optional
        }

        faidx = fa.with_suffix(fa.suffix + ".fai")
        dict_path = ref_dir / (fa.stem + ".dict")
        mmi = ref_dir / (fa.stem + ".mmi")
        bt2_prefix = ref_dir / (fa.stem + "_bt2")

        status = ReferenceIndexStatus(
            fasta=str(fa),
            reference_dir=str(ref_dir),
            faidx=str(faidx) if faidx.exists() else None,
            dict=str(dict_path) if dict_path.exists() else None,
            mmi=str(mmi) if mmi.exists() else None,
            bt2_prefix=str(bt2_prefix) if self._bowtie2_index_exists(bt2_prefix) else None,
            tools=tools,
            ok=False,
        )
        status.ok = bool(status.faidx and (status.mmi or status.bt2_prefix))
        return status

    # ---------------------------- helpers ---------------------------- #

    def _default_reference_dir_for(self, fasta: Path) -> str:
        # Prefer project workspace if provided; otherwise sibling "reference" next to the FASTA.
        if self.workspace:
            return str(self.workspace / "reference")
        return str(fasta.parent / "reference")

    def _run(self, cmd: list[str], tag: str = "cmd") -> None:
        self.log(f"[REF] $ {' '.join(cmd)}")
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
        if proc.stdout:
            self.log(proc.stdout.strip())
        if proc.returncode != 0:
            raise RuntimeError(f"[REF] {tag} failed with exit code {proc.returncode}")

    def _tool_path(self, name: str) -> Optional[str]:
        path = resolve_tool(name)
        return str(path) if path else None

    def _tool_info(self, name: str) -> Dict[str, Optional[str]]:
        path = self._tool_path(name)
        version = None
        if path:
            version = self._probe_version(name, path)
        return {"path": path, "version": version}

    def _probe_version(self, name: str, path: str) -> Optional[str]:
        try:
            if name == "samtools":
                out = subprocess.run([path, "--version"], stdout=subprocess.PIPE, text=True).stdout.strip()
                return out.splitlines()[0] if out else None
            if name == "minimap2":
                out = subprocess.run([path, "--version"], stdout=subprocess.PIPE, text=True).stdout.strip()
                return out or None
            if name == "bowtie2-build":
                out = subprocess.run([path, "--version"], stdout=subprocess.PIPE, text=True).stdout.strip()
                return out or None
            if name == "picard":
                out = subprocess.run([path], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True).stdout.strip()
                # Picard typically prints usage with version on first line
                return out.splitlines()[0] if out else None
        except Exception:
            return None
        return None

    def _supports_samtools_dict(self, samtools_path: str) -> bool:
        try:
            out = subprocess.run([samtools_path, "dict", "-h"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            return out.returncode == 0
        except Exception:
            return False

    def _try_picard_dict(self, fa_dst: Path, dict_path: Path) -> None:
        picard = self._tool_path("picard")
        if not picard:
            self.log("[REF] Picard not found; skipping .dict creation")
            return
        try:
            self._run([picard, "CreateSequenceDictionary", f"R={fa_dst}", f"O={dict_path}"], tag="picard")
        except Exception as e:
            self.log(f"[REF] Picard dict failed: {e}")

    def _bowtie2_index_exists(self, prefix: Path) -> bool:
        # Bowtie2 produces either .bt2 or .bt2l (large index) suffixes
        base = prefix.parent / prefix.name
        patterns = [
            f"{base}.1.bt2", f"{base}.2.bt2", f"{base}.3.bt2", f"{base}.4.bt2",
            f"{base}.rev.1.bt2", f"{base}.rev.2.bt2",
            f"{base}.1.bt2l", f"{base}.2.bt2l", f"{base}.3.bt2l", f"{base}.4.bt2l",
            f"{base}.rev.1.bt2l", f"{base}.rev.2.bt2l",
        ]
        return all(Path(p).exists() for p in patterns[:6]) or all(Path(p).exists() for p in patterns[6:])



import pandas as pd
import seaborn as sns
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import numpy as np

import numpy as np

from pysnptools.snpreader import Bed, Pheno
import pysnptools.util as pstutil
from pysnptools.kernelreader import SnpKernel
from pysnptools.standardizer import Unit
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sklearn.decomposition import PCA
import pandas as pd

plt.switch_backend('Agg')


class Plot:
    """Plotting class."""

    def __init__(self):
        pass

    def plot_pheno_statistics(self, pheno_file, pheno_stats_name):
        #print (pheno.iid, pheno.val)
        # Load the phenotype data from a space-delimited file
        file_path = pheno_file  # Change this to your file path

        df = pd.read_csv(file_path, delim_whitespace=True, header=None, names=['ID1', 'ID2', 'Observation'])

        # Create a PDF to save the plots
        with PdfPages(pheno_stats_name) as pdf:
            # Scatter plot: Observation vs individual
            plt.figure(figsize=(10, 6))
            plt.scatter(df.index, df['Observation'], color='#8cc5e3')
            plt.xlabel('Individual')
            plt.ylabel('Observation')
            plt.title('Observation vs Individual')
            pdf.savefig()
            plt.close()

            # Boxplot and Violin plot side by side
            plt.figure(figsize=(10, 6))

            # Boxplot
            plt.subplot(1, 2, 1)
            sns.boxplot(y=df['Observation'], color='#8cc5e3')
            plt.xlabel('Category')  # Add x-axis label
            plt.title('Boxplot of Observations')

            # Violin plot
            plt.subplot(1, 2, 2)
            sns.violinplot(y=df['Observation'], color='#b8b8b8')
            plt.xlabel('Category')  # Add x-axis label
            plt.title('Violin Plot of Observations')

            # Save the combined plot to the PDF
            pdf.savefig()
            plt.close()

            # Histogram and Density plot side by side
            plt.figure(figsize=(12, 6))

            # Histogram: Frequency vs Observation
            plt.subplot(1, 2, 1)
            plt.hist(df['Observation'], bins=30, color='#8cc5e3')
            plt.xlabel('Observation')
            plt.ylabel('Frequency')
            plt.title('Histogram: Frequency vs Observation')

            # Density plot: Density vs Observation
            plt.subplot(1, 2, 2)
            sns.kdeplot(df['Observation'], fill=True, color='#b8b8b8')
            plt.xlabel('Observation')
            plt.ylabel('Density')
            plt.title('Density Plot: Density vs Observation')

            # Save the combined plot to the PDF
            pdf.savefig()
            plt.close()

            # Cumulative density plot: Accumulative Density vs Observation
            plt.figure(figsize=(10, 6))
            sns.ecdfplot(df['Observation'], color='#8cc5e3')
            plt.xlabel('Observation')
            plt.ylabel('Cumulative Density')
            plt.title('Cumulative Density Plot: Accumulative Density vs Observation')
            pdf.savefig()
            plt.close()

    def plot_geno_statistics(self, bed, pheno, geno_stats_name):

        bed = bed.read()
        pheno = pheno.read()

        # Calculate MAF (Minor Allele Frequency)
        allele_sums = bed.val.sum(axis=0)  # Sum of allele counts for each SNP
        allele_counts = 2 * bed.iid_count  # Each individual has 2 alleles
        allele_frequencies = allele_sums / allele_counts
        maf = np.minimum(allele_frequencies, 1 - allele_frequencies)
        # Calculate heterozygosity for each individual (iid)
        heterozygosity = np.mean(bed.val == 1, axis=1)
        # Calculate heterozygosity for each marker (sid)
        marker_heterozygosity = np.mean(bed.val == 1, axis=0)

        # Perform PCA
        bed_standardized = bed.standardize(Unit())
        pca = PCA(n_components=10)
        pca_result = pca.fit_transform(bed_standardized.val)
        # Explained variance ratio
        explained_variance_ratio = pca.explained_variance_ratio_

        # Kinship
        kernel2 = bed_standardized.val @ bed_standardized.val.T / bed_standardized.sid_count
        sample_names = [''.join(iid) for iid in bed.iid[:, 1]]

        # Hertitabilty
        # Add principal components as covariates
        X_with_pcs = np.hstack((pca_result, np.ones((pca_result.shape[0], 1))))  # Include an intercept

        # Adjust phenotype for population structure
        pheno = pheno.read()
        y = pheno.val.flatten()
        y_adjusted = y - X_with_pcs @ np.linalg.lstsq(X_with_pcs, y, rcond=None)[0]

        # Calculate genetic variance (Vg) and residual variance (Ve)
        Vg = np.var(kernel2 @ y_adjusted)
        Ve = np.var(y_adjusted - kernel2 @ y_adjusted)

        # Calculate heritability (h2)
        heritability = Vg / (Vg + Ve)
        heritability_percent = heritability * 100

        # Create a PDF to save the plots
        with PdfPages(geno_stats_name) as pdf:
            plt.figure(figsize=(18, 6))  # Adjust the figure size as needed
            plt.suptitle('Marker density\n', fontsize=16)

            # Histogram of Minor Allele Frequency (MAF)
            plt.subplot(1, 3, 1)
            sns.histplot(maf, bins=9, kde=False, color='#8cc5e3', edgecolor=None)
            plt.xlabel('Minor Allele Frequency (MAF)')
            plt.ylabel('Frequency')
            plt.title('Histogram of Minor Allele Frequency (MAF)')

            # Histogram of Individual Heterozygosity
            plt.subplot(1, 3, 2)
            sns.histplot(heterozygosity, bins=5, kde=False, color='#b8b8b8', edgecolor=None)
            plt.xlabel('Individual Heterozygosity')
            plt.ylabel('Frequency')
            plt.title('Histogram of Individual Heterozygosity')

            # Histogram of Marker Heterozygosity
            plt.subplot(1, 3, 3)
            sns.histplot(marker_heterozygosity, bins=10, kde=False, color='#8cc5e3', edgecolor=None)
            plt.xlabel('Marker Heterozygosity')
            plt.ylabel('Frequency')
            plt.title('Histogram of Marker Heterozygosity')

            # Save the combined plots to the PDF
            pdf.savefig()
            plt.close()

            # Create a figure for PCA scatter plots
            plt.figure(figsize=(18, 12))  # Adjusted figure size to fit all plots on one page
            plt.suptitle('Principal component analysis', fontsize=16)
            # Define the hex color codes
            color1 = '#b8b8b8'
            color2 = '#8cc5e3'

            plot_number = 1
            for i in range(4):
                for j in range(i + 1, 4):
                    plt.subplot(2, 3, plot_number)
                    # Alternate colors based on the index
                    color = color1 if plot_number % 2 == 0 else color2
                    sns.scatterplot(x=pca_result[:, i], y=pca_result[:, j], edgecolor=None, s=50, color=color)
                    plt.xlabel(f'PC{i + 1}')
                    plt.ylabel(f'PC{j + 1}')
                    plt.title(f'PC{i + 1} vs PC{j + 1}')
                    plot_number += 1

            pdf.savefig()
            plt.close()

            plt.figure(figsize=(12, 4))
            plt.suptitle('Heritability and Explained Variance Ratio', fontsize=16)
            plt.subplots_adjust(top=0.75, wspace=0.5)

            # Heritability percentage barchart
            plt.subplot(1, 2, 1)
            plt.bar(['Heritability'], [heritability_percent], color='#8cc5e3')
            plt.ylim(0, 100)
            #plt.xlim(0, 2)
            plt.ylabel('Percentage')
            plt.title('Heritability (%) ' + str(heritability_percent))

            # Explained variance ratio plot
            plt.subplot(1, 2, 2)
            plt.bar([f'PC{i + 1}' for i in range(len(explained_variance_ratio))], explained_variance_ratio,
                    color='#b8b8b8')
            plt.ylabel('Explained Variance Ratio')
            plt.title('Explained Variance Ratio by Principal Components')

            # Save the combined plots to the PDF
            pdf.savefig()
            plt.close()

            try:
                kinship_df = pd.DataFrame(kernel2, index=sample_names, columns=sample_names)
                g = sns.clustermap(kinship_df, cmap='vlag', figsize=(15, 15), xticklabels=1, yticklabels=1,
                                   method='single', metric='euclidean')

                # Adjust the title positioning
                plt.subplots_adjust(top=0.95)
                g.fig.suptitle('Hierarchically-Clustered Kinship Matrix', fontsize=16)

                # Adjust the font size for labels
                plt.setp(g.ax_heatmap.get_xticklabels(), fontsize=4, rotation=90)
                plt.setp(g.ax_heatmap.get_yticklabels(), fontsize=4, rotation=0)
                x0, _y0, _w, _h = g.cbar_pos
                g.ax_cbar.set_position([x0, 0.9, g.ax_row_dendrogram.get_position().width - 0.1, 0.05])
                # Save the heatmap to the PDF
                pdf.savefig(g.fig)
                plt.close()
            except RecursionError:
                pass

